#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest
import math
import os
import numpy
from numpy.testing import assert_equal, assert_almost_equal
import theano
from theano import tensor
from theanolm import Vocabulary
from theanolm.network import RecurrentState
from theanolm.scoring import LatticeDecoder, SLFLattice
from theanolm.scoring.lattice import Lattice

class DummyNetwork(object):
    def __init__(self, vocabulary, projection_vector):
        self.vocabulary = vocabulary
        self.input_word_ids = tensor.matrix('input_word_ids', dtype='int64')
        self.input_class_ids = tensor.matrix('input_class_ids', dtype='int64')
        self.target_class_ids = tensor.matrix('target_class_ids', dtype='int64')
        self.mask = tensor.matrix('mask', dtype='int64')
        self.is_training = tensor.scalar('is_training', dtype='int8')
        self.recurrent_state_input = [tensor.tensor3('recurrent_state_1', dtype=theano.config.floatX)]
        self.recurrent_state_output = [self.recurrent_state_input[0] + 1]
        self.recurrent_state_size = [3]
        self.projection_vector = projection_vector

    def target_probs(self):
        num_time_steps = self.input_word_ids.shape[0]
        num_sequences = self.input_word_ids.shape[1]
        result = self.projection_vector[self.input_word_ids.flatten()]
        result += self.projection_vector[self.target_class_ids.flatten()]
        result = result.reshape([num_time_steps,
                                 num_sequences],
                                ndim=2)
        return result

class DummyLatticeDecoder(LatticeDecoder):
    def __init__(self):
        self._sorted_nodes = [Lattice.Node(id) for id in range(5)]
        self._sorted_nodes[0].time = 0.0
        self._sorted_nodes[1].time = 1.0
        self._sorted_nodes[2].time = 1.0
        self._sorted_nodes[3].time = None
        self._sorted_nodes[4].time = 3.0
        self._tokens = [[LatticeDecoder.Token()],
                        [LatticeDecoder.Token()],
                        [LatticeDecoder.Token(), LatticeDecoder.Token(), LatticeDecoder.Token()],
                        [LatticeDecoder.Token()],
                        []]
        self._tokens[0][0].total_logprob = -10.0
        self._tokens[0][0].recombination_hash = 1
        self._sorted_nodes[0].best_logprob = -10.0
        self._tokens[1][0].total_logprob = -20.0
        self._tokens[1][0].recombination_hash = 1
        self._sorted_nodes[1].best_logprob = -20.0
        self._tokens[2][0].total_logprob = -30.0
        self._tokens[2][0].recombination_hash = 1
        self._tokens[2][1].total_logprob = -50.0
        self._tokens[2][1].recombination_hash = 2
        self._tokens[2][2].total_logprob = -70.0
        self._tokens[2][2].recombination_hash = 3
        self._sorted_nodes[2].best_logprob = -30.0
        self._tokens[3][0].total_logprob = -100.0
        self._tokens[3][0].recombination_hash = 1
        self._sorted_nodes[3].best_logprob = -100.0

class TestLatticeDecoder(unittest.TestCase):
    def setUp(self):
        script_path = os.path.dirname(os.path.realpath(__file__))

        vocabulary_path = os.path.join(script_path, 'vocabulary.txt')
        with open(vocabulary_path) as vocabulary_file:
            self.vocabulary = Vocabulary.from_file(vocabulary_file, 'words')

        self.sos_id = self.vocabulary.word_to_id['<s>']
        self.yksi_id = self.vocabulary.word_to_id['yksi']
        self.kaksi_id = self.vocabulary.word_to_id['kaksi']
        self.eos_id = self.vocabulary.word_to_id['</s>']

        projection_vector = tensor.zeros(shape=(self.vocabulary.num_words(),),
                                         dtype=theano.config.floatX)
        self.sos_prob = 0.1
        projection_vector = tensor.set_subtensor(projection_vector[self.sos_id], self.sos_prob)
        self.yksi_prob = 0.2
        projection_vector = tensor.set_subtensor(projection_vector[self.yksi_id], self.yksi_prob)
        self.kaksi_prob = 0.3
        projection_vector = tensor.set_subtensor(projection_vector[self.kaksi_id], self.kaksi_prob)
        self.eos_prob = 0.4
        projection_vector = tensor.set_subtensor(projection_vector[self.eos_id], self.eos_prob)
        self.network = DummyNetwork(self.vocabulary, projection_vector)

        lattice_path = os.path.join(script_path, 'lattice.slf')
        with open(lattice_path) as lattice_file:
            self.lattice = SLFLattice(lattice_file)

    def tearDown(self):
        pass

    def test_copy_token(self):
        history = [1, 2, 3]
        token1 = LatticeDecoder.Token(history)
        token2 = LatticeDecoder.Token.copy(token1)
        token2.history.append(4)
        self.assertSequenceEqual(token1.history, [1, 2, 3])
        self.assertSequenceEqual(token2.history, [1, 2, 3, 4])

    def test_recompute_hash(self):
        token1 = LatticeDecoder.Token(history=[1, 12, 203, 3004, 23455])
        token2 = LatticeDecoder.Token(history=[2, 12, 203, 3004, 23455])
        token1.recompute_hash(None)
        token2.recompute_hash(None)
        self.assertNotEqual(token1.recombination_hash, token2.recombination_hash)
        token1.recompute_hash(5)
        token2.recompute_hash(5)
        self.assertNotEqual(token1.recombination_hash, token2.recombination_hash)
        token1.recompute_hash(4)
        token2.recompute_hash(4)
        self.assertEqual(token1.recombination_hash, token2.recombination_hash)

    def test_recompute_total(self):
        token = LatticeDecoder.Token(history=[1, 2],
                                     ac_logprob=math.log(0.1),
                                     lat_lm_logprob=math.log(0.2),
                                     nn_lm_logprob=math.log(0.3))
        token.recompute_total(0.25, 1.0, 0.0, True)
        assert_almost_equal(token.lm_logprob,
                            math.log(0.25 * 0.3 + 0.75 * 0.2))
        assert_almost_equal(token.total_logprob,
                            math.log(0.1 * (0.25 * 0.3 + 0.75 * 0.2)))
        token.recompute_total(0.25, 1.0, 0.0, False)
        assert_almost_equal(token.lm_logprob,
                            0.25 * math.log(0.3) + 0.75 * math.log(0.2))
        assert_almost_equal(token.total_logprob,
                            math.log(0.1) + 0.25 * math.log(0.3) + 0.75 * math.log(0.2))
        token.recompute_total(0.25, 10.0, 0.0, True)
        assert_almost_equal(token.lm_logprob,
                            math.log(0.25 * 0.3 + 0.75 * 0.2))
        assert_almost_equal(token.total_logprob,
                            math.log(0.1) + math.log(0.25 * 0.3 + 0.75 * 0.2) * 10.0)
        token.recompute_total(0.25, 10.0, 0.0, False)
        assert_almost_equal(token.lm_logprob,
                            0.25 * math.log(0.3) + 0.75 * math.log(0.2))
        assert_almost_equal(token.total_logprob,
                            math.log(0.1) + (0.25 * math.log(0.3) + 0.75 * math.log(0.2)) * 10.0)
        token.recompute_total(0.25, 10.0, -20.0, True)
        assert_almost_equal(token.lm_logprob,
                            math.log(0.25 * 0.3 + 0.75 * 0.2))
        assert_almost_equal(token.total_logprob,
                            math.log(0.1) + math.log(0.25 * 0.3 + 0.75 * 0.2) * 10.0 - 40.0)
        token.recompute_total(0.25, 10.0, -20.0, False)
        assert_almost_equal(token.lm_logprob,
                            0.25 * math.log(0.3) + 0.75 * math.log(0.2))
        assert_almost_equal(token.total_logprob,
                            math.log(0.1) + (0.25 * math.log(0.3) + 0.75 * math.log(0.2)) * 10.0 - 40.0)

        token = LatticeDecoder.Token(history=[1, 2],
                                     ac_logprob=-1000,
                                     lat_lm_logprob=-1001,
                                     nn_lm_logprob=-1002)
        token.recompute_total(0.75, 1.0, 0.0, True)
        # ln(exp(-1000) * (0.75 * exp(-1002) + 0.25 * exp(-1001)))
        assert_almost_equal(token.total_logprob, -2001.64263, decimal=4)

    def test_append_word(self):
        decoding_options = {
            'nnlm_weight': 1.0,
            'lm_scale': 1.0,
            'wi_penalty': 0.0,
            'ignore_unk': False,
            'unk_penalty': 0.0,
            'linear_interpolation': False,
            'max_tokens_per_node': 10,
            'beam': None,
            'recombination_order': None
        }

        initial_state = RecurrentState(self.network.recurrent_state_size)
        token1 = LatticeDecoder.Token(history=[self.sos_id], state=initial_state)
        token2 = LatticeDecoder.Token(history=[self.sos_id, self.yksi_id], state=initial_state)
        decoder = LatticeDecoder(self.network, decoding_options)

        self.assertSequenceEqual(token1.history, [self.sos_id])
        self.assertSequenceEqual(token2.history, [self.sos_id, self.yksi_id])
        assert_equal(token1.state.get(0), numpy.zeros(shape=(1,1,3)).astype(theano.config.floatX))
        assert_equal(token2.state.get(0), numpy.zeros(shape=(1,1,3)).astype(theano.config.floatX))
        self.assertEqual(token1.nn_lm_logprob, 0.0)
        self.assertEqual(token2.nn_lm_logprob, 0.0)

        decoder._append_word([token1, token2], self.kaksi_id)
        self.assertSequenceEqual(token1.history, [self.sos_id, self.kaksi_id])
        self.assertSequenceEqual(token2.history, [self.sos_id, self.yksi_id, self.kaksi_id])
        assert_equal(token1.state.get(0), numpy.ones(shape=(1,1,3)).astype(theano.config.floatX))
        assert_equal(token2.state.get(0), numpy.ones(shape=(1,1,3)).astype(theano.config.floatX))
        token1_nn_lm_logprob = math.log(self.sos_prob + self.kaksi_prob)
        token2_nn_lm_logprob = math.log(self.yksi_prob + self.kaksi_prob)
        self.assertAlmostEqual(token1.nn_lm_logprob, token1_nn_lm_logprob)
        self.assertAlmostEqual(token2.nn_lm_logprob, token2_nn_lm_logprob)

        decoder._append_word([token1, token2], self.eos_id)
        self.assertSequenceEqual(token1.history, [self.sos_id, self.kaksi_id, self.eos_id])
        self.assertSequenceEqual(token2.history, [self.sos_id, self.yksi_id, self.kaksi_id, self.eos_id])
        assert_equal(token1.state.get(0), numpy.ones(shape=(1,1,3)).astype(theano.config.floatX) * 2)
        assert_equal(token2.state.get(0), numpy.ones(shape=(1,1,3)).astype(theano.config.floatX) * 2)
        token1_nn_lm_logprob += math.log(self.kaksi_prob + self.eos_prob)
        token2_nn_lm_logprob += math.log(self.kaksi_prob + self.eos_prob)
        self.assertAlmostEqual(token1.nn_lm_logprob, token1_nn_lm_logprob)
        self.assertAlmostEqual(token2.nn_lm_logprob, token2_nn_lm_logprob)

        lm_scale = 2.0
        token1.recompute_total(1.0, lm_scale, -0.01)
        token2.recompute_total(1.0, lm_scale, -0.01)
        self.assertAlmostEqual(token1.total_logprob, token1_nn_lm_logprob * lm_scale - 0.03)
        self.assertAlmostEqual(token2.total_logprob, token2_nn_lm_logprob * lm_scale - 0.04)

    def test_prune(self):
        # token recombination
        decoder = DummyLatticeDecoder()
        decoder._max_tokens_per_node = None
        decoder._beam = None
        decoder._recombination_order = 3  # Not used.
        decoder._tokens[2][0].recombination_hash = 2
        decoder._tokens[2][1].recombination_hash = 2
        decoder._tokens[2][2].recombination_hash = 3
        decoder._prune(decoder._sorted_nodes[2])
        self.assertEqual(len(decoder._tokens[2]), 2)
        decoder._tokens[2][0].recombination_hash = 4
        decoder._tokens[2][1].recombination_hash = 4
        decoder._prune(decoder._sorted_nodes[2])
        self.assertEqual(len(decoder._tokens[2]), 1)

        # beam pruning
        decoder = DummyLatticeDecoder()
        decoder._max_tokens_per_node = None
        decoder._recombination_order = None
        # best_logprob = -20
        decoder._beam = 60
        decoder._prune(decoder._sorted_nodes[2])
        self.assertEqual(len(decoder._tokens[2]), 3)
        self.assertEqual(decoder._tokens[2][0].total_logprob, -30)
        self.assertEqual(decoder._tokens[2][1].total_logprob, -50)
        self.assertEqual(decoder._tokens[2][2].total_logprob, -70)
        decoder._beam = 50
        decoder._prune(decoder._sorted_nodes[2])
        self.assertEqual(len(decoder._tokens[2]), 2)
        self.assertEqual(decoder._tokens[2][0].total_logprob, -30)
        self.assertEqual(decoder._tokens[2][1].total_logprob, -50)
        decoder._beam = 31
        decoder._prune(decoder._sorted_nodes[2])
        self.assertEqual(len(decoder._tokens[2]), 2)
        self.assertEqual(decoder._tokens[2][0].total_logprob, -30)
        self.assertEqual(decoder._tokens[2][1].total_logprob, -50)
        decoder._beam = 15
        decoder._prune(decoder._sorted_nodes[2])
        self.assertEqual(len(decoder._tokens[2]), 1)
        self.assertEqual(decoder._tokens[2][0].total_logprob, -30)
        decoder._beam = 0
        decoder._prune(decoder._sorted_nodes[2])
        self.assertEqual(len(decoder._tokens[2]), 1)
        self.assertEqual(decoder._tokens[2][0].total_logprob, -30)

        # max tokens per node
        decoder = DummyLatticeDecoder()
        decoder._beam = None
        decoder._recombination_order = None
        decoder._max_tokens_per_node = 3
        decoder._prune(decoder._sorted_nodes[2])
        self.assertEqual(len(decoder._tokens[2]), 3)
        self.assertEqual(decoder._tokens[2][0].total_logprob, -30)
        self.assertEqual(decoder._tokens[2][1].total_logprob, -50)
        self.assertEqual(decoder._tokens[2][2].total_logprob, -70)
        decoder._max_tokens_per_node = 2
        decoder._prune(decoder._sorted_nodes[2])
        self.assertEqual(len(decoder._tokens[2]), 2)
        self.assertEqual(decoder._tokens[2][0].total_logprob, -30)
        self.assertEqual(decoder._tokens[2][1].total_logprob, -50)
        decoder._max_tokens_per_node = 1
        decoder._prune(decoder._sorted_nodes[2])
        self.assertEqual(len(decoder._tokens[2]), 1)
        self.assertEqual(decoder._tokens[2][0].total_logprob, -30)

    def test_decode(self):
        vocabulary = Vocabulary.from_word_counts({
            'TO': 1,
            'AND': 1,
            'IT': 1,
            'BUT': 1,
            'A.': 1,
            'IN': 1,
            'A': 1,
            'AT': 1,
            'THE': 1,
            'E.': 1,
            "DIDN'T": 1,
            'ELABORATE': 1})
        projection_vector = tensor.ones(shape=(vocabulary.num_words(),),
                                        dtype=theano.config.floatX)
        projection_vector *= 0.05
        network = DummyNetwork(vocabulary, projection_vector)

        decoding_options = {
            'nnlm_weight': 0.0,
            'lm_scale': None,
            'wi_penalty': None,
            'ignore_unk': False,
            'unk_penalty': None,
            'linear_interpolation': True,
            'max_tokens_per_node': None,
            'beam': None,
            'recombination_order': None
        }
        decoder = LatticeDecoder(network, decoding_options)
        tokens = decoder.decode(self.lattice)

        # Compare tokens to n-best list given by SRILM lattice-tool.
        log_scale = math.log(10)

        print()
        for token in tokens:
            print(token.ac_logprob / log_scale,
                  token.lat_lm_logprob / log_scale,
                  token.total_logprob / log_scale,
                  ' '.join(vocabulary.id_to_word[token.history]))

        all_paths = ["<s> IT DIDN'T ELABORATE </s>",
                     "<s> BUT IT DIDN'T ELABORATE </s>",
                     "<s> THE DIDN'T ELABORATE </s>",
                     "<s> AND IT DIDN'T ELABORATE </s>",
                     "<s> E. DIDN'T ELABORATE </s>",
                     "<s> IN IT DIDN'T ELABORATE </s>",
                     "<s> A DIDN'T ELABORATE </s>",
                     "<s> AT IT DIDN'T ELABORATE </s>",
                     "<s> IT IT DIDN'T ELABORATE </s>",
                     "<s> TO IT DIDN'T ELABORATE </s>",
                     "<s> A. IT DIDN'T ELABORATE </s>",
                     "<s> A IT DIDN'T ELABORATE </s>"]
        paths = [' '.join(vocabulary.id_to_word[token.history])
                 for token in tokens]
        self.assertListEqual(paths, all_paths)

        token = tokens[0]
        history = ' '.join(vocabulary.id_to_word[token.history])
        self.assertAlmostEqual(token.ac_logprob / log_scale, -8686.28, places=2)
        self.assertAlmostEqual(token.lat_lm_logprob / log_scale, -94.3896, places=2)
        self.assertAlmostEqual(token.nn_lm_logprob, math.log(0.1) * 4)

        token = tokens[1]
        self.assertAlmostEqual(token.ac_logprob / log_scale, -8743.96, places=2)
        self.assertAlmostEqual(token.lat_lm_logprob / log_scale, -111.488, places=2)
        self.assertAlmostEqual(token.nn_lm_logprob, math.log(0.1) * 5)

        token = tokens[-1]
        self.assertAlmostEqual(token.ac_logprob / log_scale, -8696.26, places=2)
        self.assertAlmostEqual(token.lat_lm_logprob / log_scale, -178.00, places=2)
        self.assertAlmostEqual(token.nn_lm_logprob, math.log(0.1) * 5)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy
from scipy.sparse import dok_matrix

class WordStatistics(object):
    """Word Unigram and Bigram Counts
    """

    def __init__(self, input_files, vocabulary=None, count_type='int32'):
        """Reads word statistics from corpus file.

        :type input_files: list of file or mmap objects
        :param input_files: input text files

        :type vocabulary: theanolm.Vocabulary
        :param vocabulary: restrict to these words
        """

        vocabulary_size = vocabulary.num_words()
        sos_id = vocabulary.word_to_id['<s>']
        eos_id = vocabulary.word_to_id['</s>']
        unk_id = vocabulary.word_to_id['<unk>']

        self.unigram_counts = numpy.zeros(vocabulary_size, count_type)
        self.bigram_counts = dok_matrix(
            (vocabulary_size, vocabulary_size), dtype=count_type)

        for subset_file in input_files:
            for line in subset_file:
                sentence = [sos_id]
                for word in line.split():
                    if word in vocabulary:
                        sentence.append(vocabulary.word_to_id[word])
                    else:
                        sentence.append(unk_id)
                sentence.append(eos_id)
                for word_id in sentence:
                    self.unigram_counts[word_id] += 1
                for left_word_id, right_word_id in zip(sentence[:-1],
                                                       sentence[1:]):
                    self.bigram_counts[left_word_id,right_word_id] += 1



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from collections import OrderedDict
import numpy
import theano
import theano.tensor as tensor
from theanolm.network.basiclayer import BasicLayer

class SamplingOutputLayer(BasicLayer):
    """Sampling Support for Output Layer

    Base class for output layers with support for sampling noise words and
    computing unnormalized probabilities. This is needed for Noise-contrastive
    estimation and BlackOut.
    """

    def _get_unnormalized_logprobs(self, layer_input):
        """Creates tensor variable that computes unnormalized output
        (preactivations).

        :type layer_input: TensorVariable
        :param layer_input: a 3-dimensional tensor that contains the input
                            vector for each time step in each sequence

        Unnormalized probabilities and noise probabilities are used by noise-
        contrastive estimation. Softmax output is exponential, so the
        preactivations can be seen as unnormalized log probabilities.
        """

        num_time_steps = layer_input.shape[0]
        num_sequences = layer_input.shape[1]
        target_class_ids = self.network.target_class_ids[:, :, None]
        result = self._get_target_preact(layer_input, target_class_ids)
        return result.reshape([num_time_steps, num_sequences])

    def _get_sample_tensors(self, layer_input):
        """Creates tensor variables for sampling k unique noise words per
        mini-batch element for NCE and BlackOut.

        :type layer_input: TensorVariable
        :param layer_input: a 3-dimensional tensor that contains the input
                            vector for each time step in each sequence

        :rtype: tuple of two TensorVariables
        :returns: 3-dimensional tensors that contain the k sampled class IDs and
                  their log probabilities for each time step in each sequence
        """

        num_time_steps = layer_input.shape[0]
        num_sequences = layer_input.shape[1]
        num_samples = self.network.num_noise_samples
        num_classes = numpy.int64(self.network.vocabulary.num_classes())
        random = self.network.random

        minibatch_size = num_time_steps * num_sequences
        if self.network.noise_probs is None:
            # The upper bound is exclusive, so this always creates samples that
            # are < num_classes.
            num_batch_samples = minibatch_size * num_samples
            sample = random.uniform((num_batch_samples,)) * num_classes
            sample = sample.astype('int64')
        else:
            # We repeat the distribution for each mini-batch element, and sample
            # k noise words per mini-batch element. k < number of outpus, so
            # it's possible without replacement.
            class_probs = self.network.noise_probs[None, :]
            class_probs = tensor.tile(class_probs, [minibatch_size, 1])
            # Since we sample different noise words for different data words, we
            # could set the probability of the correct data words to zero, as
            # suggested in the BlackOut paper. That seems to result in a little
            # bit worse model with NCE and BlackOut.
#            target_class_ids = self.network.target_class_ids.flatten()
#            target_sample_ids = tensor.arange(minibatch_size)
#            class_probs = tensor.set_subtensor(
#                class_probs[(target_sample_ids, target_class_ids)], 0)
#            denominators = class_probs.sum(1)
#            denominators = denominators[:, None]
#            class_probs /= denominators
            sample = random.multinomial_wo_replacement(pvals=class_probs,
                                                       n=num_samples)
            # For some reason (maybe a rounding error) it may happen that the
            # sample contains a very high or negative value.
            sample = tensor.maximum(sample, 0)
            sample = tensor.minimum(sample, num_classes - 1)

        sample = sample.reshape([num_time_steps, num_sequences, num_samples])
        return sample, self._get_target_preact(layer_input, sample)

    def _get_seqshared_sample_tensors(self, layer_input):
        """Creates tensor variables for sampling noise for NCE and BlackOut.
        Creates k samples for each time step. These are shared across the
        sequences in the mini-batch.

        :type layer_input: TensorVariable
        :param layer_input: a 3-dimensional tensor that contains the input
                            vector for each time step in each sequence

        :rtype: tuple of two TensorVariables
        :returns: a 2-dimensional tensor that contains k sampled class IDs for
                  each time step, and a 3-dimensional tensors that contains
                  their log probabilities for each time step in each sequence
        """

        num_time_steps = layer_input.shape[0]
        num_samples = self.network.num_noise_samples
        num_batch_samples = num_time_steps * num_samples
        num_classes = numpy.int64(self.network.vocabulary.num_classes())
        random = self.network.random

        if self.network.noise_probs is None:
            # The upper bound is exclusive, so this always creates samples that
            # are < num_classes.
            sample = random.uniform((num_batch_samples,)) * num_classes
            sample = sample.astype('int64')
        else:
            class_probs = self.network.noise_probs[None, :]
            # We could repeat the distribution for each time step, and sample k
            # noise words per time step. Since k < number of outpus, we would
            # never have a problem with sampling without replacement.
            # Unfortunately that is very inefficient.
#            class_probs = tensor.tile(class_probs, [num_time_steps, 1])
#            sample = random.multinomial_wo_replacement(pvals=class_probs,
#                                                       n=num_samples)
            sample = random.multinomial_wo_replacement(pvals=class_probs,
                                                       n=num_batch_samples)
            sample.reshape([num_time_steps, num_samples])
            # For some reason (maybe a rounding error) it may happen that the
            # sample contains a very high or negative value.
            sample = tensor.maximum(sample, 0)
            sample = tensor.minimum(sample, num_classes - 1)

        sample = sample.reshape([num_time_steps, num_samples])
        return sample, self._get_target_seq_preact(layer_input, sample)

    def _get_shared_sample_tensors(self, layer_input):
        """Creates tensor variables for sampling noise for NCE and BlackOut.
        Creates k samples for each time step. These are shared across the
        sequences in the mini-batch.

        :type layer_input: TensorVariable
        :param layer_input: a 3-dimensional tensor that contains the input
                            vector for each time step in each sequence

        :rtype: tuple of two TensorVariables
        :returns: k sampled class IDs and a 3-dimensional tensor that contain
                  their log probabilities for each time step in each sequence
        """

        num_samples = self.network.num_noise_samples
        num_classes = numpy.int64(self.network.vocabulary.num_classes())
        random = self.network.random

        # Sample k noise words in total. These are shared across mini-batch.
        if self.network.noise_probs is None:
            # The upper bound is exclusive, so this always creates samples that
            # are < num_classes
            sample = random.uniform((num_samples,)) * num_classes
            sample = sample.astype('int64')
        else:
            # Multinomial sampling is implemented for a 2-dimensional
            # distribution only.
            class_probs = self.network.noise_probs[None, :]
            sample = random.multinomial_wo_replacement(pvals=class_probs,
                                                       n=num_samples)
            sample = sample[0, :]
            # For some reason (maybe a rounding error) it may happen that the
            # sample contains a very high or negative value.
            sample = tensor.maximum(sample, 0)
            sample = tensor.minimum(sample, num_classes - 1)

        return sample, self._get_target_list_preact(layer_input, sample)

    def _get_target_preact(self, layer_input, target_class_ids):
        """Constructs the preactivations for given targets. One or more target
        outputs are given for each element in the mini-batch.

        :type layer_input: TensorVariable
        :param layer_input: a 3-dimensional tensor that contains the input
                            vector for each time step in each sequence

        :type target_class_ids: TensorVariable
        :param target_class_ids: a 3-dimensional tensor that contains one or
                                 more target class IDs for each time step in
                                 each sequence

        :rtype: TensorVariable
        :returns: a 3-dimensional tensor that contains the preactivation for
                  each target class, for each time step in each sequence
        """

        weight = self._params[self._param_path('input/W')]
        bias = self._params[self._param_path('input/b')]
        weight = weight.T
        weight = weight[target_class_ids, :]
        # The old GPU backend does not implement GpuAdvancedIncSubtensor1_dev20
        # for vectors, which is why the very slow GpuAdvancedIncSubtensor1 will
        # be selected if we index a vector.
        bias = bias[:, None]
        bias = bias[target_class_ids, 0]
#        bias = bias[target_class_ids]
        return (layer_input[:, :, None, :] * weight).sum(3) + bias

    def _get_target_seq_preact(self, layer_input, target_class_ids):
        """Constructs the preactivations for given targets. One or more target
        outputs are given for each time step in a 2-dimensional matrix.

        :type layer_input: TensorVariable
        :param layer_input: a 3-dimensional tensor that contains the input
                            vector for each time step in each sequence

        :type target_class_ids: TensorVariable
        :param target_class_ids: a 2-dimensional tensor that contains one or
                                 more target class IDs for each time step

        :rtype: TensorVariable
        :returns: a 3-dimensional tensor that contains the preactivation for
                  each target class, for each time step in each sequence
        """

        weight = self._params[self._param_path('input/W')]
        bias = self._params[self._param_path('input/b')]
        weight = weight.T
        weight = weight[target_class_ids, :]
        # The old GPU backend does not implement GpuAdvancedIncSubtensor1_dev20
        # for vectors, which is why the very slow GpuAdvancedIncSubtensor1 will
        # be selected if we index a vector.
        bias = bias[:, None]
        bias = bias[target_class_ids, 0]
#        bias = bias[target_class_ids]
        result = layer_input[:, :, None, :] * weight[:, None, :, :]
        result = result.sum(3)
        result += bias[:, None, :]
        return result

    def _get_target_list_preact(self, layer_input, target_class_ids):
        """Structures the preactivations for a list of target classes.
        Preactivations at each word are computed for all the targets.

        :type layer_input: TensorVariable
        :param layer_input: a 3-dimensional tensor that contains the input
                            vector for each time step in each sequence

        :type target_class_ids: TensorVariable
        :param target_class_ids: a list of target classes

        :rtype: TensorVariable
        :returns: a 3-dimensional tensor that contains the preactivation for
                  every target word, at each time step of each sequence
        """

        weight = self._params[self._param_path('input/W')]
        bias = self._params[self._param_path('input/b')]
        weight = weight[:, target_class_ids]
        # The old GPU backend does not implement GpuAdvancedIncSubtensor1_dev20
        # for vectors, which is why the very slow GpuAdvancedIncSubtensor1 will
        # be selected if we index a vector.
        bias = bias[:, None]
        bias = bias[target_class_ids, 0]
#        bias = bias[target_class_ids]
        return tensor.dot(layer_input, weight) + bias



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest
from os import path
import numpy
from numpy.testing import assert_almost_equal, assert_equal
import h5py
from theanolm import Vocabulary

class TestVocabulary(unittest.TestCase):
    def setUp(self):
        script_path = path.dirname(path.realpath(__file__))
        sentences1_path = path.join(script_path, 'sentences1.txt')
        sentences2_path = path.join(script_path, 'sentences2.txt')
        vocabulary_path = path.join(script_path, 'vocabulary.txt')
        classes_path = path.join(script_path, 'classes.txt')

        self.sentences1_file = open(sentences1_path)
        self.sentences2_file = open(sentences2_path)
        self.vocabulary_file = open(vocabulary_path)
        self.classes_file = open(classes_path)

    def tearDown(self):
        self.sentences1_file.close()
        self.sentences2_file.close()
        self.vocabulary_file.close()
        self.classes_file.close()

    def test_from_file(self):
        self.vocabulary_file.seek(0)
        vocabulary = Vocabulary.from_file(self.vocabulary_file, 'words')
        self.assertEqual(vocabulary.num_words(), 10 + 3)
        self.assertEqual(vocabulary.num_classes(), 10 + 3)

    def test_from_corpus(self):
        self.sentences1_file.seek(0)
        vocabulary = Vocabulary.from_corpus([self.sentences1_file])
        self.assertEqual(vocabulary.num_words(), 10 + 3)
        self.assertEqual(vocabulary.num_classes(), 10 + 3)

        self.sentences1_file.seek(0)
        self.sentences2_file.seek(0)
        vocabulary = Vocabulary.from_corpus([self.sentences1_file,
                                             self.sentences2_file],
                                            3)
        self.assertEqual(vocabulary.num_words(), 10 + 3)
        self.assertEqual(vocabulary.num_classes(), 3 + 3)

        sos_id = vocabulary.word_to_id['<s>']
        eos_id = vocabulary.word_to_id['</s>']
        unk_id = vocabulary.word_to_id['<unk>']
        self.assertEqual(sos_id, 10)
        self.assertEqual(eos_id, 11)
        self.assertEqual(unk_id, 12)
        self.assertEqual(vocabulary.word_id_to_class_id[sos_id], 3)
        self.assertEqual(vocabulary.word_id_to_class_id[eos_id], 4)
        self.assertEqual(vocabulary.word_id_to_class_id[unk_id], 5)
        word_ids = set()
        class_ids = set()
        for word in vocabulary.words():
            if not word.startswith('<'):
                word_id = vocabulary.word_to_id[word]
                word_ids.add(word_id)
                class_ids.add(vocabulary.word_id_to_class_id[word_id])
        self.assertEqual(word_ids, set(range(10)))
        self.assertEqual(class_ids, set(range(3)))

    def test_from_state(self):
        self.classes_file.seek(0)
        vocabulary1 = Vocabulary.from_file(self.classes_file, 'srilm-classes')
        f = h5py.File('in-memory.h5', driver='core', backing_store=False)
        vocabulary1.get_state(f)
        vocabulary2 = Vocabulary.from_state(f)
        self.assertTrue(numpy.array_equal(vocabulary1.id_to_word,
                                          vocabulary2.id_to_word))
        self.assertDictEqual(vocabulary1.word_to_id, vocabulary2.word_to_id)
        self.assertTrue(numpy.array_equal(vocabulary1.word_id_to_class_id,
                                          vocabulary2.word_id_to_class_id))
        self.assertListEqual(list(vocabulary1._word_classes),
                             list(vocabulary2._word_classes))

    def test_class_ids(self):
        self.classes_file.seek(0)
        vocabulary = Vocabulary.from_file(self.classes_file, 'srilm-classes')
        word_id = vocabulary.word_to_id['yksi']
        yksi_class_id = vocabulary.word_id_to_class_id[word_id]
        word_id = vocabulary.word_to_id['kaksi']
        kaksi_class_id = vocabulary.word_id_to_class_id[word_id]
        word_id = vocabulary.word_to_id['kolme']
        kolme_class_id = vocabulary.word_id_to_class_id[word_id]
        word_id = vocabulary.word_to_id['neljä']
        nelja_class_id = vocabulary.word_id_to_class_id[word_id]
        word_id = vocabulary.word_to_id['</s>']
        eos_class_id = vocabulary.word_id_to_class_id[word_id]
        self.assertNotEqual(yksi_class_id, kaksi_class_id)
        self.assertEqual(kolme_class_id, nelja_class_id)
        self.assertNotEqual(kolme_class_id, eos_class_id)
        self.assertEqual(kaksi_class_id, eos_class_id)

    def test_compute_probs(self):
        self.classes_file.seek(0)
        vocabulary = Vocabulary.from_file(self.classes_file, 'srilm-classes')
        vocabulary.compute_probs([self.sentences1_file, self.sentences2_file])

        word_id = vocabulary.word_to_id['yksi']
        self.assertAlmostEqual(vocabulary.get_word_prob(word_id), 1.0)
        word_id = vocabulary.word_to_id['kaksi']
        self.assertAlmostEqual(vocabulary.get_word_prob(word_id), 2.0 / 12.0)
        word_id = vocabulary.word_to_id['kolme']
        self.assertAlmostEqual(vocabulary.get_word_prob(word_id), 0.5)
        word_id = vocabulary.word_to_id['neljä']
        self.assertAlmostEqual(vocabulary.get_word_prob(word_id), 0.5)
        word_id = vocabulary.word_to_id['viisi']
        self.assertAlmostEqual(vocabulary.get_word_prob(word_id), 1.0)
        word_id = vocabulary.word_to_id['kuusi']
        self.assertAlmostEqual(vocabulary.get_word_prob(word_id), 0.25)
        word_id = vocabulary.word_to_id['seitsemän']
        self.assertAlmostEqual(vocabulary.get_word_prob(word_id), 0.25)
        word_id = vocabulary.word_to_id['kahdeksan']
        self.assertAlmostEqual(vocabulary.get_word_prob(word_id), 0.25)
        word_id = vocabulary.word_to_id['yhdeksän']
        self.assertAlmostEqual(vocabulary.get_word_prob(word_id), 0.25)
        word_id = vocabulary.word_to_id['kymmenen']
        self.assertAlmostEqual(vocabulary.get_word_prob(word_id), 1.0)
        word_id = vocabulary.word_to_id['<s>']
        self.assertAlmostEqual(vocabulary.get_word_prob(word_id), 1.0)
        word_id = vocabulary.word_to_id['</s>']
        self.assertAlmostEqual(vocabulary.get_word_prob(word_id), 10.0 / 12.0)
        word_id = vocabulary.word_to_id['<unk>']
        self.assertAlmostEqual(vocabulary.get_word_prob(word_id), 1.0)

    def test_get_class_memberships(self):
        vocabulary = Vocabulary.from_file(self.classes_file, 'srilm-classes')
        word_ids = numpy.array([vocabulary.word_to_id['yksi'],
                                vocabulary.word_to_id['kaksi'],
                                vocabulary.word_to_id['kolme'],
                                vocabulary.word_to_id['neljä'],
                                vocabulary.word_to_id['viisi'],
                                vocabulary.word_to_id['kuusi'],
                                vocabulary.word_to_id['seitsemän'],
                                vocabulary.word_to_id['kahdeksan'],
                                vocabulary.word_to_id['yhdeksän'],
                                vocabulary.word_to_id['kymmenen'],
                                vocabulary.word_to_id['<s>'],
                                vocabulary.word_to_id['</s>'],
                                vocabulary.word_to_id['<unk>']])
        class_ids, probs = vocabulary.get_class_memberships(word_ids)
        assert_equal(class_ids, vocabulary.word_id_to_class_id[word_ids])
        assert_equal(class_ids[3], vocabulary.word_id_to_class_id[word_ids[3]])
        assert_almost_equal(probs, [1.0,
                                    0.999,
                                    0.599 / (0.599 + 0.400),
                                    0.400 / (0.599 + 0.400),
                                    1.0,
                                    0.281 / (0.281 + 0.226 + 0.262 + 0.228),
                                    0.226 / (0.281 + 0.226 + 0.262 + 0.228),
                                    0.262 / (0.281 + 0.226 + 0.262 + 0.228),
                                    0.228 / (0.281 + 0.226 + 0.262 + 0.228),
                                    1.0,
                                    1.0,
                                    0.001,
                                    1.0])

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import os
import subprocess
import numpy
import h5py
import theano
from theanolm import Vocabulary, Architecture, Network
from theanolm.parsing import LinearBatchIterator, utterance_from_line
from theanolm.scoring import TextScorer
from theanolm.filetypes import TextFileType

def add_arguments(parser):
    argument_group = parser.add_argument_group("files")
    argument_group.add_argument(
        'model_path', metavar='MODEL-FILE', type=str,
        help='the model file that will be used to score text')
    argument_group.add_argument(
        'input_file', metavar='TEXT-FILE', type=TextFileType('r'),
        help='text file containing text to be scored (UTF-8, one sentence per '
             'line, assumed to be compressed if the name ends in ".gz")')
    argument_group.add_argument(
        '--output-file', metavar='FILE', type=TextFileType('w'), default='-',
        help='where to write the statistics (default stdout, will be '
             'compressed if the name ends in ".gz")')

    argument_group = parser.add_argument_group("scoring")
    argument_group.add_argument(
        '--output', metavar='DETAIL', type=str, default='perplexity',
        help='what to output, one of "perplexity", "utterance-scores", '
             '"word-scores" (default "perplexity")')
    argument_group.add_argument(
        '--log-base', metavar='B', type=int, default=None,
        help='convert output log probabilities to base B (default is the '
             'natural logarithm)')
    argument_group.add_argument(
        '--unk-penalty', metavar='LOGPROB', type=float, default=None,
        help="if LOGPROB is zero, do not include <unk> tokens in perplexity "
             "computation; otherwise use constant LOGPROB as <unk> token score "
             "(default is to use the network to predict <unk> probability)")

def score(args):
    with h5py.File(args.model_path, 'r') as state:
        print("Reading vocabulary from network state.")
        sys.stdout.flush()
        vocabulary = Vocabulary.from_state(state)
        print("Number of words in vocabulary:", vocabulary.num_words())
        print("Number of word classes:", vocabulary.num_classes())
        print("Building neural network.")
        sys.stdout.flush()
        architecture = Architecture.from_state(state)
        network = Network(architecture, vocabulary)
        print("Restoring neural network state.")
        sys.stdout.flush()
        network.set_state(state)

    print("Building text scorer.")
    sys.stdout.flush()
    if args.unk_penalty is None:
        ignore_unk = False
        unk_penalty = None
    elif args.unk_penalty == 0:
        ignore_unk = True
        unk_penalty = None
    else:
        ignore_unk = False
        unk_penalty = args.unk_penalty
    scorer = TextScorer(network, ignore_unk, unk_penalty)

    print("Scoring text.")
    if args.output == 'perplexity':
        _score_text(args.input_file, vocabulary, scorer, args.output_file,
                    args.log_base, False)
    elif args.output == 'word-scores':
        _score_text(args.input_file, vocabulary, scorer, args.output_file,
                    args.log_base, True)
    elif args.output == 'utterance-scores':
        _score_utterances(args.input_file, vocabulary, scorer, args.output_file,
                          args.log_base)
    else:
        print("Invalid output format requested:", args.output)
        sys.exit(1)

def _score_text(input_file, vocabulary, scorer, output_file,
                log_base=None, word_level=False):
    """Reads text from ``input_file``, computes perplexity using
    ``scorer``, and writes to ``output_file``.

    :type input_file: file object
    :param input_file: a file that contains the input sentences in SRILM n-best
                       format

    :type vocabulary: Vocabulary
    :param vocabulary: vocabulary that provides mapping between words and word
                       IDs

    :type scorer: TextScorer
    :param scorer: a text scorer for rescoring the input sentences

    :type output_file: file object
    :param output_file: a file where to write the output n-best list in SRILM
                        format

    :type log_base: int
    :param log_base: if set to other than None, convert log probabilities to
                     this base

    :type word_level: bool
    :param word_level: if set to True, also writes word-level statistics
    """

    validation_iter = \
        LinearBatchIterator(input_file,
                            vocabulary,
                            batch_size=16,
                            max_sequence_length=None)
    log_scale = 1.0 if log_base is None else numpy.log(log_base)
    unk_id = vocabulary.word_to_id['<unk>']

    total_logprob = 0.0
    num_sentences = 0
    num_words = 0
    num_probs = 0
    for word_ids, _, mask in validation_iter:
        class_ids, membership_probs = vocabulary.get_class_memberships(word_ids)
        logprobs = scorer.score_batch(word_ids, class_ids, membership_probs,
                                      mask)
        for seq_index, seq_logprobs in enumerate(logprobs):
            seq_logprob = sum(seq_logprobs)
            num_probs += len(seq_logprobs)
            total_logprob += seq_logprob
            seq_word_ids = word_ids[:, seq_index]
            seq_mask = mask[:, seq_index]
            seq_word_ids = seq_word_ids[seq_mask == 1]
            num_words += len(seq_word_ids)
            num_sentences += 1
            if not word_level:
                continue

            seq_logprobs = [x / log_scale for x in seq_logprobs]
            seq_words = vocabulary.id_to_word[seq_word_ids]
            output_file.write("# Sentence {0}\n".format(num_sentences))

            # In case some word IDs are ignored, seq_word_ids may contain more
            # items than seq_logprobs.
            logprob_index = 0
            for word_index, word_id in enumerate(seq_word_ids[1:]):
                if word_index - 2 > 0:
                    history = ['...']
                    history.extend(seq_words[word_index - 2:word_index + 1])
                else:
                    history = seq_words[0:word_index + 1]
                history = ' '.join(history)
                predicted = seq_words[word_index + 1]

                if scorer.unk_ignored() and word_id == unk_id:
                    output_file.write("p({0} | {1}) is not predicted\n".format(
                        predicted, history))
                else:
                    logprob = seq_logprobs[logprob_index]
                    logprob_index += 1
                    output_file.write("log(p({0} | {1})) = {2}\n".format(
                        predicted, history, logprob))
            assert logprob_index == len(seq_logprobs)

            # seq_logprob is in natural base.
            output_file.write("Sentence perplexity: {0}\n\n".format(
                numpy.exp(-seq_logprob / len(seq_logprobs))))

    output_file.write("Number of sentences: {0}\n".format(num_sentences))
    output_file.write("Number of words: {0}\n".format(num_words))
    output_file.write("Number of predicted probabilities: {0}\n".format(num_probs))
    if num_words > 0:
        cross_entropy = -total_logprob / num_probs
        perplexity = numpy.exp(cross_entropy)
        output_file.write("Cross entropy (base e): {0}\n".format(cross_entropy))
        if not log_base is None:
            cross_entropy /= log_scale
            output_file.write("Cross entropy (base {1}): {0}\n".format(
                cross_entropy, log_base))
        output_file.write("Perplexity: {0}\n".format(perplexity))

def _score_utterances(input_file, vocabulary, scorer, output_file,
                      log_base=None):
    """Reads utterances from ``input_file``, computes LM scores using
    ``scorer``, and writes one score per line to ``output_file``.

    Start-of-sentence and end-of-sentece tags (``<s>`` and ``</s>``) will be
    inserted at the beginning and the end of each utterance, if they're missing.
    Empty lines will be ignored, instead of interpreting them as the empty
    sentence ``<s> </s>``.

    :type input_file: file object
    :param input_file: a file that contains the input sentences in SRILM n-best
                       format

    :type vocabulary: Vocabulary
    :param vocabulary: vocabulary that provides mapping between words and word
                       IDs

    :type scorer: TextScorer
    :param scorer: a text scorer for rescoring the input sentences

    :type output_file: file object
    :param output_file: a file where to write the output n-best list in SRILM
                        format

    :type log_base: int
    :param log_base: if set to other than None, convert log probabilities to
                     this base
    """

    log_scale = 1.0 if log_base is None else numpy.log(log_base)

    unk_id = vocabulary.word_to_id['<unk>']
    num_words = 0
    num_unks = 0
    for line_num, line in enumerate(input_file):
        words = utterance_from_line(line)
        if not words:
            continue

        word_ids = vocabulary.words_to_ids(words)
        num_words += word_ids.size
        num_unks += numpy.count_nonzero(word_ids == unk_id)
        class_ids = [vocabulary.word_id_to_class_id[word_id]
                     for word_id in word_ids]
        probs = [vocabulary.get_word_prob(word_id)
                 for word_id in word_ids]

        lm_score = scorer.score_sequence(word_ids, class_ids, probs)
        lm_score /= log_scale
        output_file.write(str(lm_score) + '\n')

        if (line_num + 1) % 1000 == 0:
            print("{0} sentences scored.".format(line_num + 1))
        sys.stdout.flush()

    if num_words == 0:
        print("The input file contains no words.")
    else:
        print("{0} words processed, including start-of-sentence and "
              "end-of-sentence tags, and {1} ({2:.1f} %) out-of-vocabulary "
              "words".format(num_words, num_unks, num_unks / num_words))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from enum import Enum, unique
from collections import OrderedDict
import logging
import numpy
import theano
import theano.tensor as tensor
from theano.sandbox.rng_mrg import MRG_RandomStreams as RandomStreams
from theanolm.exceptions import IncompatibleStateError, InputError
from theanolm.network.networkinput import NetworkInput
from theanolm.network.projectionlayer import ProjectionLayer
from theanolm.network.tanhlayer import TanhLayer
from theanolm.network.grulayer import GRULayer
from theanolm.network.lstmlayer import LSTMLayer
from theanolm.network.highwaytanhlayer import HighwayTanhLayer
from theanolm.network.softmaxlayer import SoftmaxLayer
from theanolm.network.hsoftmaxlayer import HSoftmaxLayer
from theanolm.network.dropoutlayer import DropoutLayer
from theanolm.matrixfunctions import test_value

def create_layer(layer_options, *args, **kwargs):
    """Constructs one of the Layer classes based on a layer definition.

    :type layer_type: str
    :param layer_type: a text string describing the layer type
    """

    layer_type = layer_options['type']
    if layer_type == 'projection':
        return ProjectionLayer(layer_options, *args, **kwargs)
    elif layer_type == 'tanh':
        return TanhLayer(layer_options, *args, **kwargs)
    elif layer_type == 'lstm':
        return LSTMLayer(layer_options, *args, **kwargs)
    elif layer_type == 'gru':
        return GRULayer(layer_options, *args, **kwargs)
    elif layer_type == 'highwaytanh':
        return HighwayTanhLayer(layer_options, *args, **kwargs)
    elif layer_type == 'softmax':
        return SoftmaxLayer(layer_options, *args, **kwargs)
    elif layer_type == 'hsoftmax':
        return HSoftmaxLayer(layer_options, *args, **kwargs)
    elif layer_type == 'dropout':
        return DropoutLayer(layer_options, *args, **kwargs)
    else:
        raise ValueError("Invalid layer type requested: " + layer_type)

class Network(object):
    """Neural Network

    A class that creates the actual neural network graph using Theano. Functions
    that train and apply the neural network can be created by passing the input
    and output variables to ``theano.function()``.
    """

    class Mode():
        """Network Mode Selection

        Enumeration of options for selecting network mode. This will create a
        slightly different output for different purposes.

          - ``minibatch``: Process mini-batches with multiple sequences and time
                           steps. The output is a matrix with one less time
                           steps containing the probabilities of the words at
                           the next time step.
        """
        def __init__(self, minibatch=True, nce=False):
            self.minibatch = minibatch
            self.nce = nce

    def __init__(self, architecture, vocabulary, class_prior_probs=None,
                 noise_dampening=1.0, mode=None, default_device=None,
                 profile=False):
        """Initializes the neural network parameters for all layers, and
        creates Theano shared variables from them.

        When using a sampling based output layer, it needs to know the prior
        distribution of the classes, and how many noise classes to sample. The
        number of noise classes per training word is controlled by the
        ``num_noise_samples`` tensor variable. The prior distribution is a
        shared variable, so that we don't have to pass the vector to every call
        of a Theano function. The constructor initializes it using
        ``class_prior_probs`` and ``noise_dampening``.

        :type architecture: Architecture
        :param architecture: an object that describes the network architecture

        :type vocabulary: Vocabulary
        :param vocabulary: mapping between word IDs and word classes

        :type class_prior_probs: numpy.ndarray
        :param class_prior_probs: empirical (unigram) distribution of the output
                                  classes (only required for training)

        :type noise_dampening: float
        :param noise_dampening: exponent to which the unigram distribution is
                                raised before sampling noise samples

        :type mode: Network.Mode
        :param mode: selects mini-batch or single time step processing

        :type default_device: str
        :param default_device: default device where to store the shared variables

        :type profile: bool
        :param profile: if set to True, creates a Theano profile object
        """

        self.vocabulary = vocabulary
        self.architecture = architecture
        self.mode = self.Mode() if mode is None else mode

        M1 = 2147483647
        M2 = 2147462579
        random_seed = [
            numpy.random.randint(0, M1),
            numpy.random.randint(0, M1),
            numpy.random.randint(1, M1),
            numpy.random.randint(0, M2),
            numpy.random.randint(0, M2),
            numpy.random.randint(1, M2)]
        self.random = RandomStreams(random_seed)

        # Word and class inputs will be available to NetworkInput layers.
        self.input_word_ids = tensor.matrix('network/input_word_ids', dtype='int64')
        self.input_class_ids = tensor.matrix('network/input_class_ids', dtype='int64')
        if self.mode.minibatch:
            self.input_word_ids.tag.test_value = test_value(
                size=(100, 16), high=vocabulary.num_words())
            self.input_class_ids.tag.test_value = test_value(
                size=(100, 16), high=vocabulary.num_classes())
        else:
            self.input_word_ids.tag.test_value = test_value(
                size=(1, 16), high=vocabulary.num_words())
            self.input_class_ids.tag.test_value = test_value(
                size=(1, 16), high=vocabulary.num_classes())

        # During training, the output layer bias vector is initialized to the
        # unigram probabilities.
        self.class_prior_probs = class_prior_probs

        # Recurrent layers will create these lists, used to initialize state
        # variables of appropriate sizes, for doing forward passes one step at a
        # time.
        self.recurrent_state_input = []
        self.recurrent_state_size = []

        # Create the layers.
        logging.debug("Creating layers.")
        self.layers = OrderedDict()
        for input_options in architecture.inputs:
            input = NetworkInput(input_options, self)
            self.layers[input.name] = input
        for layer_description in architecture.layers:
            layer_options = self._layer_options_from_description(
                layer_description)
            if layer_options['name'] == architecture.output_layer:
                layer_options['size'] = vocabulary.num_classes()
            layer = create_layer(layer_options, self, profile=profile)
            self.layers[layer.name] = layer
        self.output_layer = self.layers[architecture.output_layer]

        # This list will be filled by the recurrent layers to contain the
        # recurrent state outputs, for doing forward passes one step at a time.
        self.recurrent_state_output = [None] * len(self.recurrent_state_size)

        # This input variable can be used to specify the classes whose
        # probabilities will be computed, instead of the whole distribution.
        self.target_class_ids = tensor.matrix('network/target_class_ids',
                                              dtype='int64')
        if self.mode.minibatch:
            self.target_class_ids.tag.test_value = test_value(
                size=(100, 16), high=vocabulary.num_classes())
        else:
            self.target_class_ids.tag.test_value = test_value(
                size=(1, 16), high=vocabulary.num_classes())

        # This input variable is used only for detecting <unk> target words.
        self.target_word_ids = tensor.matrix('network/target_word_ids',
                                             dtype='int64')
        if self.mode.minibatch:
            self.target_word_ids.tag.test_value = test_value(
                size=(100, 16), high=vocabulary.num_words())
        else:
            self.target_word_ids.tag.test_value = test_value(
                size=(1, 16), high=vocabulary.num_words())

        # Create initial parameter values.
        logging.debug("Initializing parameters.")
        self.param_init_values = OrderedDict()
        num_params = 0
        for layer in self.layers.values():
            for name, value in layer.param_init_values.items():
                logging.debug("- %s size=%d type=%s", name, value.size,
                              value.dtype)
                assert value.dtype != 'float64'
                num_params += value.size
            self.param_init_values.update(layer.param_init_values)
        logging.debug("Total number of parameters: %d", num_params)

        # Create Theano shared variables.
        if default_device is None:
            self.params = {name: theano.shared(value, name)
                           for name, value in self.param_init_values.items()}
        else:
            self.params = {name: theano.shared(value, name, target=default_device)
                           for name, value in self.param_init_values.items()}
        for layer in self.layers.values():
            layer.set_params(self.params)

        # mask is used to mask out the rest of the input matrix, when a sequence
        # is shorter than the maximum sequence length. The mask is kept as int8
        # data type, which is how Tensor stores booleans.
        if self.mode.minibatch:
            self.mask = tensor.matrix('network/mask', dtype='int8')
            self.mask.tag.test_value = test_value(size=(100, 16), high=True)
        else:
            self.mask = tensor.ones(self.input_word_ids.shape, dtype='int8')

        # Dropout layer needs to know whether we are training or evaluating.
        self.is_training = tensor.scalar('network/is_training', dtype='int8')
        self.is_training.tag.test_value = 1

        # num_noise_samples tells sampling based methods how many noise classes
        # to sample.
        self.num_noise_samples = tensor.scalar('network/num_noise_samples',
                                               dtype='int64')
        self.num_noise_samples.tag.test_value = 25

        # Sampling based methods use this noise distribution, if it's set.
        # Otherwise noise is sampled from uniform distribution.
        if (class_prior_probs is None) or (noise_dampening == 0.0):
            # Use uniform() for sampling based training.
            self.noise_probs = None
        else:
            noise_probs = numpy.power(class_prior_probs, noise_dampening)
            noise_probs /= noise_probs.sum()
            if default_device is None:
                self.noise_probs = \
                    theano.shared(noise_probs.astype(theano.config.floatX),
                                  'network/noise_probs')
            else:
                self.noise_probs = \
                    theano.shared(noise_probs.astype(theano.config.floatX),
                                  'network/noise_probs',
                                  target=default_device)

        for layer in self.layers.values():
            layer.create_structure()

    def get_state(self, state):
        """Pulls parameter values from Theano shared variables.

        If there already is a parameter in the state, it will be replaced, so it
        has to have the same number of elements.

        :type state: h5py.File
        :param state: HDF5 file for storing the neural network parameters
        """

        for name, param in self.params.items():
            if name in state:
                state[name][:] = param.get_value()
            else:
                state.create_dataset(name, data=param.get_value())

        self.architecture.get_state(state)

    def set_state(self, state):
        """Sets the values of Theano shared variables.

        Requires that ``state`` contains values for all the neural network
        parameters.

        :type state: h5py.File
        :param state: HDF5 file that contains the neural network parameters
        """

        for name, param in self.params.items():
            if not name in state:
                raise IncompatibleStateError(
                    "Parameter %s is missing from neural network state." % name)
            new_value = state[name].value
            param.set_value(new_value)
            if len(new_value.shape) == 0:
                logging.debug("%s <- %s", name, str(new_value))
            else:
                logging.debug("%s <- array%s", name, str(new_value.shape))
        try:
            self.architecture.check_state(state)
        except IncompatibleStateError as error:
            raise IncompatibleStateError(
                "Attempting to restore state of a network that is incompatible "
                "with this architecture. " + str(error))

    def add_recurrent_state(self, size):
        """Adds a recurrent state variable and returns its index.

        Used by recurrent layers to add a state variable that has to be passed
        from one time step to the next, when generating text or computing
        lattice probabilities.

        :type size: int
        :param size: size of the state vector

        :rtype size: int
        :param size: index of the new recurrent state variable
        """

        index = len(self.recurrent_state_size)
        assert index == len(self.recurrent_state_input)

        # The variables are in the structure of a mini-batch (3-dimensional
        # array) to keep the layer functions general.
        variable = tensor.tensor3('network/recurrent_state_' + str(index),
                                  dtype=theano.config.floatX)
        variable.tag.test_value = test_value(size=(1, 16, size), high=1.0)

        self.recurrent_state_size.append(size)
        self.recurrent_state_input.append(variable)

        return index

    def output_probs(self):
        """Returns the output probabilities for the whole vocabulary.

        Only computed when target_class_ids is not given.

        :rtype: TensorVariable
        :returns: a symbolic 3-dimensional matrix that contains a probability
                  for each time step, each sequence, and each output class
        """

        if not hasattr(self.output_layer, 'output_probs'):
            raise RuntimeError("The final layer is not an output layer.")
        if self.output_layer.output_probs is None:
            raise RuntimeError("Trying to read output distribution, while the "
                               "output layer has produced only target class "
                               "probabilities.")
        return self.output_layer.output_probs

    def target_probs(self):
        """Returns the output probabilities for the predicted words.

        Only computed when target_class_ids is given.

        :rtype: TensorVariable
        :returns: a symbolic 2-dimensional matrix that contains the target word
                  probability for each time step and each sequence
        """

        if not hasattr(self.output_layer, 'target_probs'):
            raise RuntimeError("The final layer is not an output layer.")
        if self.output_layer.target_probs is None:
            raise RuntimeError("Trying to read target class probabilities, "
                               "while the output layer has produced the "
                               "distribution.")
        return self.output_layer.target_probs

    def unnormalized_logprobs(self):
        """Returns the unnormalized log probabilities for the predicted words.

        These are the preactivations of the output layer, before softmax. As the
        softmax output is exponential, these can be seen as the unnormalized log
        probabilities.

        Only computed when target_class_ids is given and using softmax output.

        :rtype: TensorVariable
        :returns: a symbolic 2-dimensional matrix that contains the unnormalized
                  target word probability for each time step and each sequence
        """

        if not hasattr(self.output_layer, 'unnormalized_logprobs'):
            raise RuntimeError("The final layer is not a softmax layer, and "
                               "unnormalized probabilities are needed.")
        if self.output_layer.unnormalized_logprobs is None:
            raise RuntimeError("Trying to read target class probabilities, "
                               "while the output layer has produced the "
                               "distribution.")
        return self.output_layer.unnormalized_logprobs

    def noise_sample(self, sharing=None):
        """Returns the classes sampled from a noise distribution, and their log
        probabilities.

        Only computed when target_class_ids is given and using softmax output.

        If ``sharing`` is None``, the sample is a 3-dimensional matrix of k
        class IDs for each time step and sequence. If ``sharing`` is 'seq', the
        sample is a 2-dimensional matrix, k class IDs for each time step. If
        ``sharing`` is 'batch', the sample is a vector of k class IDs in total.
        The log probabilities are always returned in a 3-dimensional matrix, as
        they differ for each time step and sequence.

        :type sharing: str
        :param sharing: either ``None`` for k samples per mini-batch element,
                        'seq' for k samples per time step, or 'batch' for k
                        samples in total

        :rtype: tuple of two TensorVariables
        :returns: noise class IDs and their log probabilities
        """

        try:
            if sharing is None:
                return self.output_layer.sample, \
                       self.output_layer.sample_logprobs
            elif sharing == 'seq':
                return self.output_layer.seqshared_sample, \
                       self.output_layer.seqshared_sample_logprobs
            elif sharing == 'batch':
                return self.output_layer.shared_sample, \
                       self.output_layer.shared_sample_logprobs
            else:
                raise ValueError("Unknown noise sample sharing: `{}'".format(
                                 sharing))
        except AttributeError:
            raise RuntimeError(
                "Trying to read the noise sample when the final layer is not a "
                "softmax layer, or the output layer is producing a "
                "distribution instead of target probabilities.")

    def _layer_options_from_description(self, description):
        """Creates layer options based on textual architecture description.

        Most of the fields in a layer description are kept as strings. The field
        ``input_layers`` is converted to a list of actual layers found from
        ``self.layers``.

        :type description: dict
        :param description: dictionary of textual layer fields

        :rtype: dict
        :result: layer options
        """

        result = dict()
        for variable, value in description.items():
            if variable == 'inputs':
                try:
                    result['input_layers'] = [self.layers[x] for x in value]
                except KeyError as e:
                    raise InputError("Input layer `{}' does not exist, when "
                                     "creating layer `{}'.".format(
                                     e.args[0],
                                     description['name']))
            else:
                result[variable] = value
        return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Reads a class definitions file and an n-gram counts file, and corrects the
# class expansion probabilities according to the unigram counts of the words.

import argparse
import sys
from filetypes import TextFileType
from wordclasses import WordClasses, WordsToClasses
from ngramcounts import NGramCounts

parser = argparse.ArgumentParser()
parser.add_argument('classes', type=TextFileType('r'), help='input class definitions file')
parser.add_argument('counts', type=TextFileType('r'), help='n-gram counts file')
args = parser.parse_args()

classes = WordClasses()
classes.read(args.classes)

word_counts = NGramCounts()
word_counts.read(args.counts)

for cls in classes:
	counts = dict()
	for word, prob in cls:
		unigram = tuple([word])
		if unigram in word_counts:
			counts[word] = word_counts[unigram]
		else:
			counts[word] = 0
	total_count = sum(counts.values())
	if total_count != 0:
		for word, count in counts.items():
			cls.set_probability(word, float(count) / total_count)

classes.write(sys.stdout)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# A Python class that stores word class definitions.

class WordClasses:
	class WordClass:
		def __init__(self, name):
			self.__name = name
			self.__probs = dict()
		
		def __contains__(self, word):
			return word in self.__probs
		
		def __iter__(self):
			for word, prob in self.__probs.items():
				yield word, prob
		
		def name(self):
			return self.__name
		
		def add(self, word, prob):
			self.__probs[word] = float(prob)
		
		def get_probability(self, word):
			return self.__probs[word]
		
		def set_probability(self, word, prob):
			if not word in self.__probs:
				raise ValueError("WordClasses.set_probability: No word " + word + " in class " + self.__name + ".")
			self.__probs[word] = prob
		
		def normalize(self):
			total_prob = sum(self.__probs.values())
			if total_prob > 0:
				factor = 1 / total_prob
			else:
				factor = 0
			for word, prob in self.__probs.items():
				self.__probs[word] = prob * factor
		
		def write(self, output_file):
			for word, prob in self.__probs.items():
				output_file.write(self.__name)
				output_file.write(' ')
				output_file.write(str(prob))
				output_file.write(' ')
				output_file.write(word)
				output_file.write('\n')
	
	def __init__(self):
		self.__classes = dict()
		self.__next_id = 1
	
	def __iter__(self):
		for name, cls in self.__classes.items():
			yield cls
	
	def read(self, input_file):
		for line in input_file:
			fields = line.split()
			if len(fields) == 0:
				continue
			elif len(fields) == 2:
				# Assume mkcls / word2vec file format.
				word = fields[0]
				name = "CLASS-" + fields[1].zfill(5)
				prob = 0
			elif len(fields) == 3:
				# Assume SRILM file format.
				name = fields[0]
				prob = float(fields[1])
				word = fields[2]
			else:
				raise Exception("Invalid word class definition: " + line)
			if not name in self.__classes:
				self.__classes[name] = self.WordClass(name)
			self.__classes[name].add(word, prob)

	def write(self, output_file):
		for name, cls in self.__classes.items():
			cls.normalize()
			cls.write(output_file)
	
	def create(self, name=None):
		if name is None:
			while True:
				name = 'CLASS-' + ('%05d' % self.__next_id)
				self.__next_id += 1
				if not name in self.__classes.keys():
					break
		new_class = self.WordClass(name)
		self.__classes[name] = new_class
		return new_class
	
	def find_containing(self, word):
		for name, cls in self.__classes.items():
			if word in cls:
				return cls
		return None

class WordsToClasses:
	def __init__(self, word_classes):
		self.__map = dict()
		for cls in word_classes:
			for word, prob in cls:
				self.__map[word] = cls.name()
	
	def __getitem__(self, word):
		return self.__map[word]
	
	def __contains__(self, word):
		return word in self.__map



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy
import theano
import theano.tensor as tensor
from theanolm.matrixfunctions import get_submatrix
from theanolm.network.basiclayer import BasicLayer

class LSTMLayer(BasicLayer):
    """Long Short-Term Memory Layer

    A. Graves, J. Schmidhuber (2005)
    Framewise phoneme classification with bidirectional LSTM and other neural
    network architectures
    Neural Networks, 18(5–6), 602–610
    """

    def __init__(self, *args, **kwargs):
        """Initializes the parameters used by this layer.

        The weight matrices are concatenated so that they can be applied in a
        single parallel matrix operation. The same thing for bias vectors.
        Input, forget, and output gate biases are initialized to -1.0, 1.0, and
        -1.0 respectively, so that in the beginning of training, the forget gate
        activation will be almost 1.0 (meaning that the LSTM does not default to
        forgetting information).
        """

        super().__init__(*args, **kwargs)

        input_size = sum(x.output_size for x in self.input_layers)
        output_size = self.output_size

        # Add state variables to be passed between time steps.
        self.cell_state_index = self.network.add_recurrent_state(output_size)
        self.hidden_state_index = self.network.add_recurrent_state(output_size)

        # Initialize the parameters.
        num_gates = 3
        # layer input weights for each gate and the candidate state
        self._init_orthogonal_weight('layer_input/W', input_size, output_size,
                                     scale=0.01, count=num_gates+1)
        # hidden state input weights for each gate and the candidate state
        self._init_orthogonal_weight('step_input/W', output_size, output_size,
                                     count=num_gates+1)
        # biases for each gate and the candidate state
        self._init_bias('layer_input/b', output_size, [-1.0, 1.0, -1.0, 0.0])

    def create_structure(self):
        """Creates the symbolic graph of this layer.

        The input is always 3-dimensional: the first dimension is the time step,
        the second dimension are the sequences, and the third dimension is the
        layer input. When processing mini-batches, all dimensions can have size
        greater than one.

        The function can also be used to create a structure for generating the
        probability distribution of the next word. Then the input is still
        3-dimensional, but the size of the first dimension (time steps) is 1,
        and the state outputs from the previous time step are read from
        ``self.network.recurrent_state_input``.

        Saves the recurrent state in the Network object: cell state C_(t) and
        hidden state h_(t). ``self.output`` will be set to the hidden state
        output, which is the actual output of this layer.

        Assumes that the shared variables have been passed using
        ``set_params()``.
        """

        layer_input = tensor.concatenate([x.output for x in self.input_layers],
                                         axis=2)
        num_time_steps = layer_input.shape[0]
        num_sequences = layer_input.shape[1]

        # Compute the gate and candidate state pre-activations, which don't
        # depend on the state input from the previous time step.
        layer_input_preact = self._tensor_preact(layer_input, 'layer_input')

        # Weights of the hidden state input of each time step have to be applied
        # inside the loop.
        hidden_state_weights = self._get_param('step_input/W')

        if self.network.mode.minibatch:
            sequences = [self.network.mask, layer_input_preact]
            non_sequences = [hidden_state_weights]
            initial_cell_state = tensor.zeros(
                (num_sequences, self.output_size), dtype=theano.config.floatX)
            initial_hidden_state = tensor.zeros(
                (num_sequences, self.output_size), dtype=theano.config.floatX)

            state_outputs, _ = theano.scan(
                self._create_time_step,
                sequences=sequences,
                outputs_info=[initial_cell_state, initial_hidden_state],
                non_sequences=non_sequences,
                name='lstm_steps',
                n_steps=num_time_steps,
                profile=self._profile,
                strict=True)

            self.output = state_outputs[1]
        else:
            cell_state_input = \
                self.network.recurrent_state_input[self.cell_state_index]
            hidden_state_input = \
                self.network.recurrent_state_input[self.hidden_state_index]

            state_outputs = self._create_time_step(
                mask[0],
                layer_input_preact[0],
                cell_state_input[0],
                hidden_state_input[0],
                hidden_state_weights)

            # Create a new axis for time step with size 1.
            cell_state_output = cell_state_output[None,:,:]
            hidden_state_output = hidden_state_output[None,:,:]

            self.network.recurrent_state_output[self.cell_state_index] = \
                cell_state_output
            self.network.recurrent_state_output[self.hidden_state_index] = \
                hidden_state_output
            self.output = hidden_state_output

    def _create_time_step(self, mask, x_preact, C_in, h_in, h_weights):
        """The LSTM step function for theano.scan(). Creates the structure of
        one time step.

        The inputs do not contain the time step dimension. ``mask`` is a vector
        containing a boolean mask for each sequence. ``x_preact`` is a matrix
        containing the preactivations for each sequence. ``C_in`` and ``h_in``,
        as well as the outputs, are matrices containing the state vectors for
        each sequence.

        The required affine transformations have already been applied to the
        input prior to creating the loop. The transformed inputs and the mask
        that will be passed to the step function are vectors when processing a
        mini-batch - each value corresponds to the same time step in a different
        sequence.

        :type mask: TensorVariable
        :param mask: a symbolic vector that masks out sequences that are past
                     the last word

        :type x_preact: TensorVariable
        :param x_preact: concatenation of the input x_(t) pre-activations
                         computed using the gate and candidate state weights and
                         biases; shape is (the number of sequences, state size *
                         4)

        :type C_in: TensorVariable
        :param C_in: C_(t-1), cell state output of the previous time step; shape
                     is (the number of sequences, state size)

        :type h_in: TensorVariable
        :param h_in: h_(t-1), hidden state output of the previous time step;
                     shape is (the number of sequences, state size)

        :type h_weights: TensorVariable
        :param h_weights: concatenation of the gate and candidate state weights
                          to be applied to h_(t-1); shape is (state size, state
                          size * 4)

        :rtype: a tuple of two TensorVariables
        :returns: C_(t) and h_(t), the cell state and hidden state outputs
        """

        # pre-activation of the gates and candidate state
        preact = tensor.dot(h_in, h_weights)
        preact += x_preact

        # input, forget, and output gates
        i = tensor.nnet.sigmoid(get_submatrix(preact, 0, self.output_size))
        f = tensor.nnet.sigmoid(get_submatrix(preact, 1, self.output_size))
        o = tensor.nnet.sigmoid(get_submatrix(preact, 2, self.output_size))

        # cell state and hidden state outputs
        C_candidate = tensor.tanh(get_submatrix(preact, 3, self.output_size))
        C_out = f * C_in + i * C_candidate
        h_out = o * tensor.tanh(C_out)

        # Apply the mask. None creates a new axis with size 1, causing the mask
        # to be broadcast to all the outputs.
        C_out = tensor.switch(mask[:,None], C_out, C_in)
        h_out = tensor.switch(mask[:,None], h_out, h_in)

        return C_out, h_out



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from collections import OrderedDict
import logging
import numpy
import theano
import theano.tensor as tensor
from theanolm.network.basiclayer import BasicLayer

class DropoutLayer(BasicLayer):
    """Dropout Layer

    A dropout layer is not a regular layer in the sense that it doesn't contain
    any neurons. It simply randomly sets some activations to zero at train time
    to prevent overfitting.

    N. Srivastava et al. (2014)
    Dropout: A Simple Way to Prevent Neural Networks from Overfitting
    Journal of Machine Learning Research, 15, 1929-1958
    """

    def __init__(self, layer_options, *args, **kwargs):
        """Validates the parameters of this layer.
        """

        super().__init__(layer_options, *args, **kwargs)

        if 'dropout_rate' in layer_options:
            self.dropout_rate = float(layer_options['dropout_rate'])
            # Don't allow dropout rate too close to 1.0 to prevent division by
            # zero.
            self.dropout_rate = min(0.9999, self.dropout_rate)
        else:
            self.dropout_rate = 0.5
        logging.debug("  dropout_rate=%f", self.dropout_rate)

        # Make sure the user hasn't tried to change the number of connections.
        input_size = sum(x.output_size for x in self.input_layers)
        output_size = self.output_size
        if input_size != output_size:
            raise ValueError("Dropout layer cannot change the number of "
                             "connections.")

    def create_structure(self):
        """Creates the symbolic graph of this layer.

        Sets ``self.output`` to a symbolic matrix that describes the output of
        this layer. During training sets randomly some of the outputs to zero.
        Assumes that the shared variables have been passed using
        ``set_params()``.
        """

        layer_input = tensor.concatenate([x.output for x in self.input_layers],
                                         axis=2)
        # Pass rate is the probability of not dropping a unit.
        pass_rate = 1.0 - self.dropout_rate
        sample = self.network.random.uniform(size=layer_input.shape)
        mask = tensor.cast(sample < pass_rate, theano.config.floatX)
        # Multiply the output by the inverse of the pass rate before dropping
        # units to compensate the scaling effect.
        scale_correction = 1.0 / pass_rate
        self.output = tensor.switch(self.network.is_training,
                                    layer_input * scale_correction * mask,
                                    layer_input)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import logging
from time import time
import h5py
import numpy
import theano
from theanolm import ShufflingBatchIterator, LinearBatchIterator
from theanolm.exceptions import IncompatibleStateError, NumberError
from theanolm.training.stoppers import create_stopper

class Trainer(object):
    """Training Process
    
    Saves a history of validation costs and decreases learning rate when the
    cost does not decrease anymore.
    """

    def __init__(self, training_options, vocabulary, training_files, sampling):
        """Creates the optimizer and initializes the training process.

        Creates empty member variables for the perplexities list and the
        training state at the validation point. Training state is saved at only
        one validation point at a time, so validation interval is at least the
        the number of samples used per validation.

        :type training_options: dict
        :param training_options: a dictionary of training options

        :type vocabulary: Vocabulary
        :param vocabulary: vocabulary that provides mapping between words and
                           word IDs

        :type training_files: list of file objects
        :param training_files: list of files to be used as training data

        :type sampling: list of floats
        :param sampling: specifies a fraction for each training file, how much
                         to sample on each epoch
        """

        self._vocabulary = vocabulary

        print("Computing unigram probabilities and the number of mini-batches "
              "in training data.")
        linear_iter = LinearBatchIterator(
            training_files,
            vocabulary,
            batch_size=training_options['batch_size'],
            max_sequence_length=training_options['sequence_length'])
        sys.stdout.flush()
        self._updates_per_epoch = 0
        class_counts = numpy.zeros(vocabulary.num_classes(), dtype='int64')
        for word_ids, _, mask in linear_iter:
            self._updates_per_epoch += 1
            word_ids = word_ids[mask == 1]
            class_ids = vocabulary.word_id_to_class_id[word_ids]
            numpy.add.at(class_counts, class_ids, 1)
        if self._updates_per_epoch < 1:
            raise ValueError("Training data does not contain any sentences.")
        logging.debug("One epoch of training data contains %d mini-batch updates.",
                      self._updates_per_epoch)
        self.class_prior_probs = class_counts / class_counts.sum()
        logging.debug("Class unigram probabilities are in the range [%.8f, "
                      "%.8f].",
                      self.class_prior_probs.min(),
                      self.class_prior_probs.max())

        self._training_iter = ShufflingBatchIterator(
            training_files,
            sampling,
            vocabulary,
            batch_size=training_options['batch_size'],
            max_sequence_length=training_options['sequence_length'])

        self._stopper = create_stopper(training_options, self)
        self._options = training_options

        # iterator to cross-validation data, or None for no cross-validation
        self._validation_iter = None
        # a text scorer for performing cross-validation
        self._scorer = None
        # number of perplexity samples per validation
        self._samples_per_validation = 7
        # function for combining validation samples
        self._statistic_function = lambda x: numpy.median(numpy.asarray(x))
        # the stored validation samples
        self._local_perplexities = []
        # the state at the center of validation samples
        self._validation_state = None

        # number of mini-batch updates between log messages
        self._log_update_interval = 0

        # the network to be trained
        self._network = None
        # the optimization function
        self._optimizer = None
        # current candidate for the minimum validation cost state
        self._candidate_state = None

    def set_validation(self, validation_iter, scorer,
                       samples_per_validation=None, statistics_function=None):
        """Sets cross-validation iterator and parameters.

        :type validation_iter: BatchIterator
        :param validation_iter: an iterator for computing validation set
                                perplexity

        :type scorer: TextScorer
        :param scorer: a text scorer for computing validation set perplexity

        :type samples_per_validation: int
        :param samples_per_validation: number of perplexity samples to compute
                                       per cross-validation

        :type statistic_function: Python function
        :param statistic_function: a function to be performed on a list of
           consecutive perplexity measurements to compute the validation cost
           (median by default)
        """

        self._validation_iter = validation_iter
        self._scorer = scorer

        if not samples_per_validation is None:
            self._samples_per_validation = samples_per_validation

        if not statistics_function is None:
            self._statistics_function = statistics_function

    def set_logging(self, log_interval):
        """Sets logging parameters.

        :type log_interval: int
        :param log_interval: number of mini-batch updates between log messages
        """

        self._log_update_interval = log_interval

    def initialize(self, network, state, optimizer):
        """Sets the network and the HDF5 file that stores the network state,
        optimizer, and validation scorer and iterator.

        If the HDF5 file contains a network state, initializes the network with
        that state.

        :type network: Network
        :param network: the network, which will be used to retrieve state when
                        saving

        :type state: h5py.File
        :param state: HDF5 file where initial training state will be possibly
                      read from, and candidate states will be saved to

        :type optimizer: BasicOptimizer
        :param optimizer: one of the optimizer implementations
        """

        self._network = network
        self._optimizer = optimizer

        self._candidate_state = state
        if 'trainer' in self._candidate_state:
            print("Restoring initial network state from {}.".format(
                self._candidate_state.filename))
            sys.stdout.flush()
            self._reset_state()
        else:
            # index to the cost history that corresponds to the current candidate
            # state
            self._candidate_index = None
            # current training epoch
            self.epoch_number = 1
            # number of mini-batch updates performed in this epoch
            self.update_number = 0
            # validation set cost history
            self._cost_history = numpy.asarray([], dtype=theano.config.floatX)

        # total number of mini-batch updates performed (after restart)
        self._total_updates = 0

    def train(self):
        """Trains a neural network.

        If cross-validation has been configured using ``set_validation()``,
        computes the validation set perplexity as many times per epoch as
        specified by the _validation_frequency_ option and saves the model when
        the perplexity improves. Otherwise saves the model after each epoch.
        """

        if (self._network is None) or (self._optimizer is None) or \
           (self._candidate_state is None):
            raise RuntimeError("Trainer has not been initialized before "
                               "calling train().")

        start_time = time()
        while self._stopper.start_new_epoch():
            epoch_start_time = time()
            for word_ids, file_ids, mask in self._training_iter:
                self.update_number += 1
                self._total_updates += 1

                class_ids = self._vocabulary.word_id_to_class_id[word_ids]
                update_start_time = time()
                self._optimizer.update_minibatch(word_ids, class_ids, file_ids, mask)
                self._update_duration = time() - update_start_time

                if (self._log_update_interval >= 1) and \
                   (self._total_updates % self._log_update_interval == 0):
                    self._log_update()

                self._validate()

                if not self._stopper.start_new_minibatch():
                    break

            if self._validation_iter is None:
                self._set_candidate_state()

            epoch_duration = time() - epoch_start_time
            epoch_minutes = epoch_duration / 60
            epoch_time_h, epoch_time_m = divmod(epoch_minutes, 60)
            message = "Finished training epoch {} in {:.0f} hours {:.1f} minutes." \
                      .format(self.epoch_number, epoch_time_h, epoch_time_m)
            best_cost = self.candidate_cost()
            if not best_cost is None:
                message += " Best validation perplexity {:.2f}.".format(
                    best_cost)
            print(message)

            self.epoch_number += 1
            self.update_number = 0

        duration = time() - start_time
        minutes = duration / 60
        time_h, time_m = divmod(minutes, 60)
        print("Training finished in {:.0f} hours {:.1f} minutes." \
              .format(time_h, time_m))

    def get_state(self, state):
        """Pulls parameter values from Theano shared variables and updates a
        HDF5 file with all the network and training state variables.

        For consistency, all the parameter values are returned as numpy types,
        since state read from a model file also contains numpy types. This also
        ensures the cost history will be copied into the returned dictionary.

        :type state: h5py.File
        :param state: HDF5 file for storing the current state
        """

        h5_trainer = state.require_group('trainer')
        h5_trainer.attrs['epoch_number'] = self.epoch_number
        h5_trainer.attrs['update_number'] = self.update_number
        if 'cost_history' in h5_trainer:
            h5_trainer['cost_history'].resize(self._cost_history.shape)
            h5_trainer['cost_history'][:] = self._cost_history
        else:
            h5_trainer.create_dataset(
                'cost_history', data=self._cost_history, maxshape=(None,),
                chunks=(1000,))

        if not self._network is None:
            self._network.get_state(state)
        self._training_iter.get_state(state)
        if not self._optimizer is None:
            self._optimizer.get_state(state)

    def _reset_state(self):
        """Resets the values of Theano shared variables to the current candidate
         state.

        Sets candidate state index point to the last element in the loaded cost
        history.

        Requires that if ``state`` is set, it contains values for all the
        training parameters.

        :type state: h5py.File
        :param state: if a HDF5 file is given, reads the the training parameters
                      from this file, and assumes this is the state of minimum
                      cost found so far
        """

        self._network.set_state(self._candidate_state)

        if not 'trainer' in self._candidate_state:
            raise IncompatibleStateError("Training state is missing.")
        h5_trainer = self._candidate_state['trainer']

        if not 'epoch_number' in h5_trainer.attrs:
            raise IncompatibleStateError("Current epoch number is missing from "
                                         "training state.")
        self.epoch_number = int(h5_trainer.attrs['epoch_number'])

        if not 'update_number' in h5_trainer.attrs:
            raise IncompatibleStateError("Current update number is missing "
                                         "from training state.")
        self.update_number = int(h5_trainer.attrs['update_number'])

        logging.info("[%d] (%.2f %%) of epoch %d",
                     self.update_number,
                     self.update_number / self._updates_per_epoch * 100,
                     self.epoch_number)

        if 'cost_history' in h5_trainer:
            self._cost_history = h5_trainer['cost_history'].value
            if self._cost_history.size == 0:
                print("Validation set cost history is empty in the training state.")
                self._candidate_index = None
            else:
                self._candidate_index = self._cost_history.size - 1
                self._log_validation()
        else:
            print("Warning: Validation set cost history is missing from "
                  "training state. Initializing to empty cost history.")
            self._cost_history = numpy.asarray([], dtype=theano.config.floatX)
            self._candidate_index = None

        self._training_iter.set_state(self._candidate_state)
        self._optimizer.set_state(self._candidate_state)

    def num_validations(self):
        """Returns the number of validations performed.

        :rtype: int
        :returns: size of cost history
        """

        return self._cost_history.size

    def validations_since_candidate(self):
        """Returns the number of times the validation set cost has been computed
        since the current candidate for optimal state was obtained.

        :rtype: int
        :returns: number of validations since the current candidate state (0
                  means the current candidate is the last validation)
        """

        if self._cost_history.size == 0:
            raise RuntimeError("BasicTrainer.validations_since_candidate() "
                               "called with empty cost history.")

        return self._cost_history.size - 1 - self._candidate_index

    def candidate_cost(self):
        """Returns the validation set cost given by the current candidate for
        the minimum cost state.

        :rtype: float
        :returns: current candidate state cost, or None if the candidate state
                  has not been set yet
        """

        if self._candidate_index is None:
            return None

        return self._cost_history[self._candidate_index]

    def _decrease_learning_rate(self):
        """Called when the validation set cost stops decreasing.
        """

        # Current learning rate might be smaller than the one stored in the
        # state, so set the new value after restoring optimizer to the old
        # state.
        old_value = self._optimizer.learning_rate
        new_value = old_value / 2
        self._reset_state()
        self._stopper.improvement_ceased()
        self._optimizer.learning_rate = new_value

        print("Model performance stopped improving. Decreasing learning rate "
              "from {} to {} and resetting state to {:.0f} % of epoch {}."
              .format(old_value,
                      new_value,
                      self.update_number / self._updates_per_epoch * 100,
                      self.epoch_number))

    def _has_improved(self):
        """Tests whether the previously computed validation set cost was
        significantly better than the cost given by the current candidate state.

        TODO: Implement a test for statistical significance.

        :rtype: bool
        :returns: True if validation set cost decreased enough, or there was no
                  previous candidate state; False otherwise
        """

        if self._cost_history.size == 0:
            raise RuntimeError("BasicTrainer._has_improved() called with empty "
                               "cost history.")

        if self._candidate_index is None:
            return True

        return self._cost_history[-1] < 0.999 * self.candidate_cost()

    def _log_update(self):
        """Logs information about the previous mini-batch update.
        """

        logging.info("[%d] (%.1f %%) of epoch %d -- lr = %.1g, cost = %.2f, "
                     "duration = %.1f ms",
                     self.update_number,
                     self.update_number / self._updates_per_epoch * 100,
                     self.epoch_number,
                     self._optimizer.learning_rate,
                     self._optimizer.update_cost,
                     self._update_duration * 100)

    def _log_validation(self):
        """Prints the validation set cost history (or its tail), highlighting
        the candidate for the minimum cost.
        """

        str_costs = ["%.1f" % x for x in self._cost_history]
        if not self._candidate_index is None:
            str_costs[self._candidate_index] = \
                '[' + str_costs[self._candidate_index] + ']'
        logging.debug("[%d] Validation set cost history: %s",
                      self.update_number,
                      ' '.join(str_costs[-20:]))

    def _set_candidate_state(self, state=None):
        """Sets neural network and training state as the candidate for the
        minimum validation cost state, and writes to disk.

        :type state: h5py.File
        :param state: if a HDF5 file is given, reads the state from this file,
                      instead of the current state
        """

        if state is None:
            self.get_state(self._candidate_state)
        else:
            state.flush()
            for name in state:
                if name in self._candidate_state:
                    del self._candidate_state[name]
                self._candidate_state.copy(state[name], name, expand_refs=True)
            for name in state.attrs:
                self._candidate_state.attrs[name] = state.attrs[name]

        if self._cost_history.size == 0:
            self._candidate_index = None
        else:
            self._candidate_index = self._cost_history.size - 1

        self._candidate_state.flush()
        logging.info("New candidate for optimal state saved to %s.",
                     self._candidate_state.filename)

    def _validate(self):
        """If at or just before the actual validation point, computes perplexity
        and adds to the list of samples. At the actual validation point we have
        `self._samples_per_validation` values and combine them using
        `self._statistic_function`. If the model performance has improved, the
        state at the center of the validation samples will be saved using
        `self._set_candidate_state()`.

        :type perplexity: float
        :param perplexity: computed perplexity at a validation point, None
                           elsewhere
        """

        if self._validation_iter is None:
            return  # Validation has not been configured.

        if not self._is_scheduled(self._options['validation_frequency'],
                                  self._samples_per_validation - 1):
            return  # We don't have to validate now.

        perplexity = self._scorer.compute_perplexity(self._validation_iter)
        if numpy.isnan(perplexity) or numpy.isinf(perplexity):
            raise NumberError("Validation set perplexity computation resulted "
                              "in a numerical error.")

        self._local_perplexities.append(perplexity)
        if len(self._local_perplexities) == 1:
            logging.debug("[%d] First validation sample, perplexity %.2f.",
                          self.update_number,
                          perplexity)

        # The rest of the function will be executed only at and after the center
        # of sampling points.
        if not self._is_scheduled(self._options['validation_frequency'],
                                  self._samples_per_validation // 2):
            return

        # The first sampling point within samples_per_validation / 2 of the
        # actual validation point is the center of the sampling points. This
        # will be saved in case the model performance has improved.
        if self._validation_state is None:
            logging.debug("[%d] Center of validation, perplexity %.2f.",
                          self.update_number,
                          perplexity)
            self._validation_state = h5py.File(
                name='validation-state', driver='core', backing_store=False)
            self.get_state(self._validation_state)

        # The rest of the function will be executed only at the final sampling
        # point.
        if not self._is_scheduled(self._options['validation_frequency']):
            return
        logging.debug("[%d] Last validation sample, perplexity %.2f.",
                      self.update_number,
                      perplexity)

        if len(self._local_perplexities) < self._samples_per_validation:
            # After restoring a previous validation state, which is at the
            # center of the sampling points, the trainer will collect again half
            # of the samples. Don't take that as a validation.
            logging.debug("[%d] Only %d samples collected. Ignoring this "
                          "validation.",
                          self.update_number,
                          len(self._local_perplexities))
            self._local_perplexities = []
            self._validation_state.close()
            self._validation_state = None
            return

        statistic = self._statistic_function(self._local_perplexities)
        self._cost_history = numpy.append(self._cost_history, statistic)
        if self._has_improved():
            # Take the state at the actual validation point and replace the cost
            # history with the current cost history that also includes this
            # latest statistic.
            h5_cost_history = self._validation_state['trainer/cost_history']
            h5_cost_history.resize(self._cost_history.shape)
            h5_cost_history[:] = self._cost_history
            self._set_candidate_state(self._validation_state)

        self._log_validation()

        if (self._options['patience'] >= 0) and \
           (self.validations_since_candidate() > self._options['patience']):
            # Too many validations without finding a new candidate state.

            # If any validations have been done, the best state has been found
            # and saved. If training has been started from previous state,
            # _candidate_state has been set to the initial state.
            assert not self._candidate_state is None

            self._decrease_learning_rate()

        self._local_perplexities = []
        self._validation_state.close()
        self._validation_state = None

    def _is_scheduled(self, frequency, within=0):
        """Checks if an event is scheduled to be performed within given number
        of updates after this point.

        For example, updates_per_epoch=9, frequency=2:

        update_number:  1   2   3   4  [5]  6   7   8  [9] 10  11  12
        * frequency:    2   4   6   8  10  12  14  16  18  20  22  24
        modulo:         2   4   6   8   1   3   5   7   0   2   4   6
        within:         4   3   2   1   0   3   2   1   0   4   3   2
        * frequency:    8   6   4   2   0   6   4   2   0   8   6   4

        :type frequency: int
        :param frequency: how many times per epoch the event should be
                          performed

        :type within: int
        :param within: if zero, returns True if the event should be performed
                       now; otherwise returns True if the event should be
                       performed within this many updates in the future

        :rtype: bool
        :returns: whether the operation is scheduled to be performed
        """

        modulo = self.update_number * frequency % self._updates_per_epoch
        return modulo < frequency or \
               self._updates_per_epoch - modulo <= within * frequency



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import mmap
import logging
import numpy
import h5py
import theano
from theanolm import Vocabulary, Architecture, Network
from theanolm import LinearBatchIterator
from theanolm.training import Trainer, create_optimizer
from theanolm.scoring import TextScorer
from theanolm.filetypes import TextFileType

def add_arguments(parser):
    argument_group = parser.add_argument_group("files")
    argument_group.add_argument(
        'model_path', metavar='MODEL-FILE', type=str,
        help='path where the best model state will be saved in HDF5 binary '
             'data format')
    argument_group.add_argument(
        '--training-set', metavar='FILE', type=TextFileType('r'), nargs='+',
        required=True,
        help='text files containing training data (UTF-8, one sentence per '
             'line, assumed to be compressed if the name ends in ".gz")')
    argument_group.add_argument(
        '--validation-file', metavar='VALID-FILE', type=TextFileType('r'),
        default=None,
        help='text file containing validation data for early stopping (UTF-8, '
             'one sentence per line, assumed to be compressed if the name ends '
             'in ".gz")')
    argument_group.add_argument(
        '--vocabulary', metavar='FILE', type=str, default=None,
        help='word or class vocabulary to be used in the neural network input '
             'and output, in the format specified by the --vocabulary-format '
             'argument (UTF-8 text, default is to use all the words from the '
             'training data)')
    argument_group.add_argument(
        '--vocabulary-format', metavar='FORMAT', type=str, default='words',
        help='format of the file specified with --vocabulary argument, one of '
             '"words" (one word per line, default), "classes" (word and class '
             'ID per line), "srilm-classes" (class name, membership '
             'probability, and word per line)')
    argument_group = parser.add_argument_group("network architecture")
    argument_group.add_argument(
        '--architecture', metavar='FILE', type=str, default='lstm300',
        help='path to neural network architecture description, or a standard '
             'architecture name, "lstm300" or "lstm1500" (default "lstm300")')
    argument_group.add_argument(
        '--num-classes', metavar='N', type=int, default=None,
        help='generate N classes using a simple word frequency based algorithm '
             'when --vocabulary argument is not given (default is to not use '
             'word classes)')

    argument_group = parser.add_argument_group("training process")
    argument_group.add_argument(
        '--sampling', metavar='FRACTION', type=float, nargs='*', default=[],
        help='randomly sample only FRACTION of each training file on each '
             'epoch (list the fractions in the same order as the training '
             'files)')
    argument_group.add_argument(
        '--sequence-length', metavar='N', type=int, default=100,
        help='ignore sentences longer than N words (default 100)')
    argument_group.add_argument(
        '--batch-size', metavar='N', type=int, default=16,
        help='each mini-batch will contain N sentences (default 16)')
    argument_group.add_argument(
        '--validation-frequency', metavar='N', type=int, default='5',
        help='cross-validate for reducing learning rate or early stopping N '
             'times per training epoch (default 5)')
    argument_group.add_argument(
        '--patience', metavar='N', type=int, default=4,
        help='allow perplexity to increase N consecutive cross-validations, '
             'before decreasing learning rate; if less than zero, never '
             'decrease learning rate (default 4)')
    argument_group.add_argument(
        '--random-seed', metavar='N', type=int, default=None,
        help='seed to initialize the random state (default is to seed from a '
             'random source provided by the oprating system)')

    argument_group = parser.add_argument_group("optimization")
    argument_group.add_argument(
        '--optimization-method', metavar='NAME', type=str, default='adagrad',
        help='optimization method, one of "sgd", "nesterov", "adagrad", '
             '"adadelta", "rmsprop-sgd", "rmsprop-nesterov", "adam" '
             '(default "adagrad")')
    argument_group.add_argument(
        '--learning-rate', metavar='ALPHA', type=float, default=0.1,
        help='initial learning rate (default 0.1)')
    argument_group.add_argument(
        '--momentum', metavar='BETA', type=float, default=0.9,
        help='momentum coefficient for momentum optimization methods (default '
             '0.9)')
    argument_group.add_argument(
        '--gradient-decay-rate', metavar='GAMMA', type=float, default=0.9,
        help='geometric rate for averaging gradients (default 0.9)')
    argument_group.add_argument(
        '--sqr-gradient-decay-rate', metavar='GAMMA', type=float, default=0.999,
        help='geometric rate for averaging squared gradients in Adam optimizer '
             '(default 0.999)')
    argument_group.add_argument(
        '--numerical-stability-term', metavar='EPSILON', type=float,
        default=1e-6,
        help='a value that is used to prevent instability when dividing by '
             'very small numbers (default 1e-6)')
    argument_group.add_argument(
        '--gradient-normalization', metavar='THRESHOLD', type=float,
        default=5,
        help='scale down the gradients if necessary to make sure their norm '
             '(normalized by mini-batch size) will not exceed THRESHOLD '
             '(default 5)')
    argument_group.add_argument(
        '--cost', metavar='NAME', type=str, default='cross-entropy',
        help='cost function, one of "cross-entropy" (default), "nce" '
             '(noise-contrastive estimation), or "blackout"')
    argument_group.add_argument(
        '--num-noise-samples', metavar='K', type=int, default=5,
        help='sampling based costs sample K noise words per one training word '
             '(default 5)')
    argument_group.add_argument(
        '--noise-sharing', metavar='SHARING', type=str, default=None,
        help='can be "seq" for sharing noise samples between mini-batch '
             'sequences, or "batch" for sharing noise samples across einter '
             'mini-batch for improved speed (default is no sharing, which is '
             'very slow)')
    argument_group.add_argument(
        '--noise-dampening', metavar='ALPHA', type=float, default=0.5,
        help='the empirical unigram distribution is raised to the power ALPHA '
             'before sampling noise words; 0.0 corresponds to the uniform '
             'distribution and 1.0 corresponds to the unigram distribution '
             '(default 0.5)')
    argument_group.add_argument(
        '--unk-penalty', metavar='LOGPROB', type=float, default=None,
        help="if LOGPROB is zero, do not include <unk> tokens in perplexity "
             "computation; otherwise use constant LOGPROB as <unk> token score "
             "(default is to use the network to predict <unk> probability)")
    argument_group.add_argument(
        '--weights', metavar='LAMBDA', type=float, nargs='*', default=[],
        help='scale a mini-batch update by LAMBDA if the data is from the '
             'corresponding training file (list the weights in the same order '
             'as the training files)')

    argument_group = parser.add_argument_group("early stopping")
    argument_group.add_argument(
        '--stopping-criterion', metavar='NAME', type=str,
        default='annealing-count',
        help='selects a criterion for early-stopping, one of "epoch-count" '
             '(fixed number of epochs), "no-improvement" (no improvement since '
             'learning rate was decreased), "annealing-count" (default, '
             'learning rate is decreased a fixed number of times)')
    argument_group.add_argument(
        '--min-epochs', metavar='N', type=int, default=1,
        help='perform at least N training epochs (default 1)')
    argument_group.add_argument(
        '--max-epochs', metavar='N', type=int, default=100,
        help='perform at most N training epochs (default 100)')
    argument_group.add_argument(
        '--max-annealing-count', metavar='N', type=int, default=0,
        help='when using annealing-count stopping criterion, continue training '
             'after decreasing learning rate at most N times (default 0)')

    argument_group = parser.add_argument_group("configuration")
    argument_group.add_argument(
        '--default-device', metavar='DEVICE', type=str, default=None,
        help='when multiple GPUs are present, use DEVICE as default')

    argument_group = parser.add_argument_group("logging and debugging")
    argument_group.add_argument(
        '--log-file', metavar='FILE', type=str, default='-',
        help='path where to write log file (default is standard output)')
    argument_group.add_argument(
        '--log-level', metavar='LEVEL', type=str, default='info',
        help='minimum level of events to log, one of "debug", "info", "warn" '
             '(default "info")')
    argument_group.add_argument(
        '--log-interval', metavar='N', type=int, default=1000,
        help='print statistics of every Nth mini-batch update; quiet if less '
             'than one (default 1000)')
    argument_group.add_argument(
        '--debug', action="store_true",
        help='use test values to get better error messages from Theano')
    argument_group.add_argument(
        '--print-graph', action="store_true",
        help='print Theano computation graph')
    argument_group.add_argument(
        '--profile', action="store_true",
        help='enable profiling Theano functions')

def train(args):
    numpy.random.seed(args.random_seed)

    log_file = args.log_file
    log_level = getattr(logging, args.log_level.upper(), None)
    if not isinstance(log_level, int):
        print("Invalid logging level requested:", args.log_level)
        sys.exit(1)
    log_format = '%(asctime)s %(funcName)s: %(message)s'
    if args.log_file == '-':
        logging.basicConfig(stream=sys.stdout, format=log_format, level=log_level)
    else:
        logging.basicConfig(filename=log_file, format=log_format, level=log_level)

    if args.debug:
        theano.config.compute_test_value = 'warn'
        print("Enabled computing test values for tensor variables.")
        print("Warning: GpuArray backend will fail random number generation!")
    else:
        theano.config.compute_test_value = 'off'
    theano.config.profile = args.profile
    theano.config.profile_memory = args.profile

    with h5py.File(args.model_path, 'a', driver='core') as state:
        if state.keys():
            print("Reading vocabulary from existing network state.")
            sys.stdout.flush()
            vocabulary = Vocabulary.from_state(state)
        elif args.vocabulary is None:
            print("Constructing vocabulary from training set.")
            sys.stdout.flush()
            vocabulary = Vocabulary.from_corpus(args.training_set,
                                                args.num_classes)
            for training_file in args.training_set:
                training_file.seek(0)
            vocabulary.get_state(state)
        else:
            print("Reading vocabulary from {}.".format(args.vocabulary))
            sys.stdout.flush()
            with open(args.vocabulary, 'rt', encoding='utf-8') as vocab_file:
                vocabulary = Vocabulary.from_file(vocab_file,
                                                  args.vocabulary_format)
                if args.vocabulary_format == 'classes':
                    print("Computing class membership probabilities from "
                          "unigram word counts.")
                    sys.stdout.flush()
                    vocabulary.compute_probs(args.training_set)
            vocabulary.get_state(state)
        print("Number of words in vocabulary:", vocabulary.num_words())
        print("Number of word classes:", vocabulary.num_classes())

        if (args.num_noise_samples > vocabulary.num_classes()):
            print("Number of noise samples ({}) is larger than the number of "
                  "classes. This doesn't make sense and would cause sampling "
                  "to fail.".format(args.num_noise_samples))
            sys.exit(1)

        if args.unk_penalty is None:
            ignore_unk = False
            unk_penalty = None
        elif args.unk_penalty == 0:
            ignore_unk = True
            unk_penalty = None
        else:
            ignore_unk = False
            unk_penalty = args.unk_penalty

        num_training_files = len(args.training_set)
        if len(args.weights) > num_training_files:
            print("You specified more weights than training files.")
            sys.exit(1)
        weights = numpy.ones(num_training_files).astype(theano.config.floatX)
        for index, weight in enumerate(args.weights):
            weights[index] = weight

        training_options = {
            'batch_size': args.batch_size,
            'sequence_length': args.sequence_length,
            'validation_frequency': args.validation_frequency,
            'patience': args.patience,
            'stopping_criterion': args.stopping_criterion,
            'max_epochs': args.max_epochs,
            'min_epochs': args.min_epochs,
            'max_annealing_count': args.max_annealing_count
        }
        logging.debug("TRAINING OPTIONS")
        for option_name, option_value in training_options.items():
            logging.debug("%s: %s", option_name, str(option_value))

        optimization_options = {
            'method': args.optimization_method,
            'epsilon': args.numerical_stability_term,
            'gradient_decay_rate': args.gradient_decay_rate,
            'sqr_gradient_decay_rate': args.sqr_gradient_decay_rate,
            'learning_rate': args.learning_rate,
            'weights': weights,
            'momentum': args.momentum,
            'max_gradient_norm': args.gradient_normalization,
            'cost_function': args.cost,
            'num_noise_samples': args.num_noise_samples,
            'noise_sharing': args.noise_sharing,
            'ignore_unk': ignore_unk,
            'unk_penalty': unk_penalty
        }
        logging.debug("OPTIMIZATION OPTIONS")
        for option_name, option_value in optimization_options.items():
            if type(option_value) is list:
                value_str = ', '.join(str(x) for x in option_value)
                logging.debug("%s: [%s]", option_name, value_str)
            else:
                logging.debug("%s: %s", option_name, str(option_value))

        if len(args.sampling) > len(args.training_set):
            print("You specified more sampling coefficients than training "
                  "files.")
            sys.exit(1)

        print("Creating trainer.")
        sys.stdout.flush()
        trainer = Trainer(training_options, vocabulary, args.training_set,
                          args.sampling)
        trainer.set_logging(args.log_interval)

        print("Building neural network.")
        sys.stdout.flush()
        if args.architecture == 'lstm300' or args.architecture == 'lstm1500':
            architecture = Architecture.from_package(args.architecture)
        else:
            with open(args.architecture, 'rt', encoding='utf-8') as arch_file:
                architecture = Architecture.from_description(arch_file)

        network = Network(architecture, vocabulary, trainer.class_prior_probs,
                          args.noise_dampening,
                          default_device=args.default_device,
                          profile=args.profile)

        print("Compiling optimization function.")
        sys.stdout.flush()
        optimizer = create_optimizer(optimization_options, network,
                                     device=args.default_device,
                                     profile=args.profile)

        if args.print_graph:
            print("Cost function computation graph:")
            theano.printing.debugprint(optimizer.gradient_update_function)

        trainer.initialize(network, state, optimizer)

        if not args.validation_file is None:
            print("Building text scorer for cross-validation.")
            sys.stdout.flush()
            scorer = TextScorer(network, ignore_unk, unk_penalty, args.profile)
            print("Validation text:", args.validation_file.name)
            validation_mmap = mmap.mmap(args.validation_file.fileno(),
                                        0,
                                        prot=mmap.PROT_READ)
            validation_iter = \
                LinearBatchIterator(validation_mmap,
                                    vocabulary,
                                    batch_size=args.batch_size,
                                    max_sequence_length=None)
            trainer.set_validation(validation_iter, scorer)
        else:
            print("Cross-validation will not be performed.")
            validation_iter = None

        print("Training neural network.")
        sys.stdout.flush()
        trainer.train()

        if not 'layers' in state.keys():
            print("The model has not been trained. No cross-validations were "
                  "performed or training did not improve the model.")
        elif not validation_iter is None:
            network.set_state(state)
            perplexity = scorer.compute_perplexity(validation_iter)
            print("Best validation set perplexity:", perplexity)



from theanolm.parsing.linearbatchiterator import LinearBatchIterator
from theanolm.parsing.shufflingbatchiterator import ShufflingBatchIterator
from theanolm.parsing.functions import utterance_from_line



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy
from shlex import shlex
from theanolm.exceptions import InputError
from theanolm.probfunctions import logprob_type
from theanolm.scoring.lattice import Lattice

class SLFLattice(Lattice):
    """SLF Format Word Lattice

    A word lattice that can be read in SLF format.
    """

    def __init__(self, lattice_file):
        """Reads an SLF lattice file.

        If ``lattice_file`` is ``None``, creates an empty lattice (useful for
        testing).

        :type lattice_file: file object
        :param lattice_file: a file in SLF lattice format
        """

        super().__init__()

        # No log conversion by default. "None" means the lattice file uses
        # linear probabilities.
        self._log_scale = logprob_type(1.0)

        self._initial_node_id = None
        self._final_node_id = None

        if lattice_file is None:
            self._num_nodes = 0
            self._num_links = 0
            return

        self._num_nodes = None
        self._num_links = None
        for line in lattice_file:
            fields = self._split_slf_line(line)
            self._read_slf_header(fields)
            if (not self._num_nodes is None) and (not self._num_links is None):
                break

        if not self.wi_penalty is None:
            if self._log_scale is None:
                self.wi_penalty = numpy.log(self.wi_penalty)
            else:
                self.wi_penalty *= self._log_scale

        self.nodes = [self.Node(id) for id in range(self._num_nodes)]

        for line in lattice_file:
            fields = self._split_slf_line(line)
            if not fields:
                continue
            name, value = self._split_slf_field(fields[0])
            if name == 'I':
                self._read_slf_node(int(value), fields[1:])
            elif name == 'J':
                self._read_slf_link(int(value), fields[1:])

        if len(self.links) != self._num_links:
            raise InputError("Number of links in SLF lattice doesn't match the "
                             "LINKS field.")

        if not self._initial_node_id is None:
            self.initial_node = self.nodes[self._initial_node_id]
        else:
            # Find the node with no incoming links.
            self.initial_node = None
            for node in self.nodes:
                if len(node.in_links) == 0:
                    self.initial_node = node
                    break
            if self.initial_node is None:
                raise InputError("Could not find initial node in SLF lattice.")

        if not self._final_node_id is None:
            self.final_node = self.nodes[self._final_node_id]
        else:
            # Find the node with no outgoing links.
            self.final_node = None
            for node in self.nodes:
                if len(node.out_links) == 0:
                    self.final_node = node
                    break
            if self.final_node is None:
                raise InputError("Could not find final node in SLF lattice.")

        # If word identity information is not present in node definitions then
        # it must appear in link definitions.
        self._move_words_to_links()
        for link in self.links:
            if link.word is None:
                raise InputError("SLF lattice does not contain word identity "
                                 "in link %d or in the following node.".format(
                                 link.id))

    def _read_slf_header(self, fields):
        """Reads SLF lattice header fields and saves them in member variables.

        :type fields: list of strs
        :param fields: fields, such as name="value"
        """
        
        for field in fields:
            name, value = self._split_slf_field(field)
            if (name == 'UTTERANCE') or (name == 'U'):
                self.utterance_id = value
            elif (name == 'SUBLAT') or (name == 'S'):
                raise InputError("Sub-lattices are not supported.")
            elif name == 'base':
                value = numpy.float64(value)
                if value == 0.0:
                    self._log_scale = None
                else:
                    self._log_scale = logprob_type(numpy.log(value))
            elif name == 'lmscale':
                self.lm_scale = logprob_type(value)
            elif name == 'wdpenalty':
                self.wi_penalty = logprob_type(value)
            elif name == 'start':
                self._initial_node_id = int(value)
            elif name == 'end':
                self._final_node_id = int(value)
            elif (name == 'NODES') or (name == 'N'):
                self._num_nodes = int(value)
            elif (name == 'LINKS') or (name == 'L'):
                self._num_links = int(value)

    def _read_slf_node(self, node_id, fields):
        """Reads SLF lattice node fields and saves the information in the given
        node.

        Some SLF lattices contain word identities in the nodes. They will be
        saved into the node, but later moved to the links leading to the node.

        :type node_id: int
        :param node_id: ID of the node

        :type fields: list of strs
        :param fields: the rest of the node fields after ID
        """

        node = self.nodes[node_id]
        for field in fields:
            name, value = self._split_slf_field(field)
            if (name == 'time') or (name == 't'):
                node.time = float(value)
            elif (name == 'WORD') or (name == 'W'):
                node.word = value

    def _read_slf_link(self, link_id, fields):
        """Reads SLF lattice link fields and creates such link.

        :type link_id: int
        :param link_id: ID of the link

        :type fields: list of strs
        :param fields: the rest of the link fields after ID
        """

        start_node = None
        end_node = None
        word = None
        ac_logprob = None
        lm_logprob = None

        for field in fields:
            name, value = self._split_slf_field(field)
            if (name == 'START') or (name == 'S'):
                start_node = self.nodes[int(value)]
            elif (name == 'END') or (name == 'E'):
                end_node = self.nodes[int(value)]
            elif (name == 'WORD') or (name == 'W'):
                word = value
            elif (name == 'acoustic') or (name == 'a'):
                if self._log_scale is None:
                    ac_logprob = logprob_type(numpy.log(numpy.float64(value)))
                else:
                    ac_logprob = logprob_type(value) * self._log_scale
            elif (name == 'language') or (name == 'l'):
                if self._log_scale is None:
                    lm_logprob = logprob_type(numpy.log(numpy.float64(value)))
                else:
                    lm_logprob = logprob_type(value) * self._log_scale

        if start_node is None:
            raise InputError("Start node is not specified for link %d.".format(
                             link_id))
        if end_node is None:
            raise InputError("End node is not specified for link %d.".format(
                             link_id))
        link = self._add_link(start_node, end_node)
        link.word = word
        link.ac_logprob = ac_logprob
        link.lm_logprob = lm_logprob

    def _split_slf_line(self, line):
        """Parses a list of fields from an SLF lattice line.

        Each field contains a name, followed by =, followed by a possible quoted
        value. Only double quotes can be used for quotation, and for the literal
        " the double quote must be escaped (\"). I'm not surprise if other
        implementations or the standard doesn't agree.

        :type line: str
        :param line: a line from an SLF file

        :rtype: list of strs
        :returns: list of fields found from the line, with possible quotation
                  marks removed
        """

        lex = shlex(line, posix=True)
        lex.quotes = '"'
        lex.wordchars += "'"
        lex.whitespace_split = True
        return list(lex)

    def _split_slf_field(self, field):
        """Parses the name and value from an SLF lattice field.

        :type field: str
        :param field: a field, such as 'UTTERANCE=utterance 123'

        :rtype: tuple of two strs
        :returns: the name and value of the field
        """
        
        name_value = field.split('=', 1)
        if len(name_value) != 2:
            raise InputError("Expected '=' in SLF lattice field: '{}'".format(
                             field))
        name = name_value[0]
        value = name_value[1]
        return name, value

    def _move_words_to_links(self):
        """Move word identities from nodes to the links leading to the node.

        SLF lattices may contain words either in the nodes or in the links. If
        they are in the nodes, move them to the link leading to the node. Note
        that if there's a word in the initial node, it will be discarded. This
        is what SRILM does as well, and the SLF format says that the words are
        in the end nodes in forward lattices.
        """

        visited = { self.initial_node.id }

        def visit_link(link):
            end_node = link.end_node
            if hasattr(end_node, 'word') and isinstance(end_node.word, str):
                if link.word is None:
                    link.word = end_node.word
                else:
                    raise InputError("SLF lattice contains words both in nodes "
                                     "and links.")
            if not end_node.id in visited:
                visited.add(end_node.id)
                for next_link in end_node.out_links:
                    visit_link(next_link)

        for link in self.initial_node.out_links:
            visit_link(link)

        for node in self.nodes:
            if hasattr(node, 'word'):
                del node.word



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest
import os
import mmap
import numpy
from numpy.testing import assert_equal
import theanolm
from theanolm.parsing.functions import find_sentence_starts

class TestIterators(unittest.TestCase):
    def setUp(self):
        script_path = os.path.dirname(os.path.realpath(__file__))
        sentences1_path = os.path.join(script_path, 'sentences1.txt')
        sentences2_path = os.path.join(script_path, 'sentences2.txt')
        vocabulary_path = os.path.join(script_path, 'vocabulary.txt')

        self.sentences1_file = open(sentences1_path)
        self.sentences2_file = open(sentences2_path)
        self.vocabulary_file = open(vocabulary_path)
        self.vocabulary = theanolm.Vocabulary.from_file(self.vocabulary_file, 'words')

    def tearDown(self):
        self.sentences1_file.close()
        self.sentences2_file.close()
        self.vocabulary_file.close()

    def test_find_sentence_starts(self):
        sentences1_mmap = mmap.mmap(self.sentences1_file.fileno(),
                                    0,
                                    access=mmap.ACCESS_READ)
        sentence_starts = find_sentence_starts(sentences1_mmap)
        self.sentences1_file.seek(sentence_starts[0])
        self.assertEqual(self.sentences1_file.readline(), 'yksi kaksi\n')
        self.sentences1_file.seek(sentence_starts[1])
        self.assertEqual(self.sentences1_file.readline(), 'kolme neljä viisi\n')
        self.sentences1_file.seek(sentence_starts[2])
        self.assertEqual(self.sentences1_file.readline(), 'kuusi seitsemän kahdeksan\n')
        self.sentences1_file.seek(sentence_starts[3])
        self.assertEqual(self.sentences1_file.readline(), 'yhdeksän\n')
        self.sentences1_file.seek(sentence_starts[4])
        self.assertEqual(self.sentences1_file.readline(), 'kymmenen\n')
        self.sentences1_file.seek(0)

        sentences2_mmap = mmap.mmap(self.sentences2_file.fileno(),
                                    0,
                                    access=mmap.ACCESS_READ)
        sentence_starts = find_sentence_starts(sentences2_mmap)
        self.sentences2_file.seek(sentence_starts[0])
        self.assertEqual(self.sentences2_file.readline(), 'kymmenen yhdeksän\n')
        self.sentences2_file.seek(sentence_starts[1])
        self.assertEqual(self.sentences2_file.readline(), 'kahdeksan seitsemän kuusi\n')
        self.sentences2_file.seek(sentence_starts[2])
        self.assertEqual(self.sentences2_file.readline(), 'viisi\n')
        self.sentences2_file.seek(sentence_starts[3])
        self.assertEqual(self.sentences2_file.readline(), 'neljä\n')
        self.sentences2_file.seek(sentence_starts[4])
        self.assertEqual(self.sentences2_file.readline(), 'kolme kaksi yksi\n')
        self.sentences2_file.seek(0)

    def test_shuffling_batch_iterator(self):
        iterator = theanolm.ShufflingBatchIterator([self.sentences1_file,
                                                    self.sentences2_file],
                                                   [],
                                                   self.vocabulary,
                                                   batch_size=2,
                                                   max_sequence_length=5)

        sentences1 = []
        files1 = []
        for word_ids, file_ids, mask in iterator:
            class_ids = self.vocabulary.word_id_to_class_id[word_ids]
            for sequence in range(2):
                sequence_mask = mask[:,sequence]
                sequence_word_ids = word_ids[sequence_mask != 0,sequence]
                sequence_class_ids = class_ids[sequence_mask != 0,sequence]
                sequence_file_ids = file_ids[sequence_mask != 0,sequence]
                assert_equal(sequence_word_ids, sequence_class_ids)
                sentences1.append(' '.join(self.vocabulary.id_to_word[sequence_word_ids]))
                files1.extend(sequence_file_ids)
        self.assertEqual(files1.count(0), 20)
        self.assertEqual(files1.count(1), 20)
        sentences1_str = ' '.join(sentences1)
        sentences1_sorted_str = ' '.join(sorted(sentences1))
        self.assertEqual(sentences1_sorted_str,
                         '<s> kahdeksan seitsemän kuusi </s> '
                         '<s> kolme kaksi yksi </s> '
                         '<s> kolme neljä viisi </s> '
                         '<s> kuusi seitsemän kahdeksan </s> '
                         '<s> kymmenen </s> '
                         '<s> kymmenen yhdeksän </s> '
                         '<s> neljä </s> '
                         '<s> viisi </s> '
                         '<s> yhdeksän </s> '
                         '<s> yksi kaksi </s>')
        self.assertEqual(len(iterator), 5)

        sentences2 = []
        files2 = []
        for word_ids, file_ids, mask in iterator:
            class_ids = self.vocabulary.word_id_to_class_id[word_ids]
            for sequence in range(2):
                sequence_mask = mask[:,sequence]
                sequence_word_ids = word_ids[sequence_mask != 0,sequence]
                sequence_class_ids = class_ids[sequence_mask != 0,sequence]
                sequence_file_ids = file_ids[sequence_mask != 0,sequence]
                assert_equal(sequence_word_ids, sequence_class_ids)
                sentences2.append(' '.join(self.vocabulary.id_to_word[sequence_word_ids]))
                files2.extend(sequence_file_ids)
        self.assertCountEqual(sentences1, sentences2)
        self.assertCountEqual(files1, files2)
        self.assertTrue(sentences1 != sentences2)
        self.assertTrue(files1 != files2)

        # The sentences are wraped so that we get more sequences if we limit
        # the maximum length. Sequences shorter than two words will be ignored.
        iterator = theanolm.ShufflingBatchIterator([self.sentences1_file,
                                                    self.sentences2_file],
                                                   [],
                                                   self.vocabulary,
                                                   batch_size=2,
                                                   max_sequence_length=4)
        self.assertEqual(len(iterator), 5)
        iterator = theanolm.ShufflingBatchIterator([self.sentences1_file,
                                                    self.sentences2_file],
                                                   [],
                                                   self.vocabulary,
                                                   batch_size=2,
                                                   max_sequence_length=3)
        self.assertEqual(len(iterator), 7)

        # Sample 2 and 4 sentences (40 % and 80 %).
        iterator = theanolm.ShufflingBatchIterator([self.sentences1_file,
                                                    self.sentences2_file],
                                                   [0.4, 0.8],
                                                   self.vocabulary,
                                                   batch_size=1,
                                                   max_sequence_length=5)
        self.assertEqual(len(iterator), 2 + 4)

        # Make sure there are no duplicates.
        self.assertSetEqual(set(iterator._order),
                            set(numpy.unique(iterator._order)))
        self.assertEqual(numpy.count_nonzero(iterator._order <= 4), 2)
        self.assertEqual(numpy.count_nonzero(iterator._order >= 5), 4)

    def test_linear_batch_iterator(self):
        iterator = theanolm.LinearBatchIterator(self.sentences1_file,
                                                self.vocabulary,
                                                batch_size=2,
                                                max_sequence_length=5)
        words = []
        for word_ids, file_ids, mask in iterator:
            class_ids = self.vocabulary.word_id_to_class_id[word_ids]
            assert_equal(word_ids, class_ids)
            assert_equal(file_ids, 0)
            for sequence in range(mask.shape[1]):
                sequence_mask = mask[:,sequence]
                sequence_word_ids = word_ids[sequence_mask != 0,sequence]
                words.extend(self.vocabulary.id_to_word[sequence_word_ids])
        self.assertEqual(' '.join(words),
                         '<s> yksi kaksi </s> '
                         '<s> kolme neljä viisi </s> '
                         '<s> kuusi seitsemän kahdeksan </s> '
                         '<s> yhdeksän </s> '
                         '<s> kymmenen </s>')

        iterator = theanolm.LinearBatchIterator([self.sentences1_file,
                                                 self.sentences2_file],
                                                self.vocabulary,
                                                batch_size=2,
                                                max_sequence_length=5)
        words = []
        all_file_ids = []
        for word_ids, file_ids, mask in iterator:
            class_ids = self.vocabulary.word_id_to_class_id[word_ids]
            assert_equal(word_ids, class_ids)
            for sequence in range(mask.shape[1]):
                sequence_mask = mask[:,sequence]
                sequence_word_ids = word_ids[sequence_mask != 0,sequence]
                words.extend(self.vocabulary.id_to_word[sequence_word_ids])
                sequence_file_ids = file_ids[sequence_mask != 0,sequence]
                all_file_ids.extend(sequence_file_ids)
        self.assertEqual(' '.join(words),
                         '<s> yksi kaksi </s> '
                         '<s> kolme neljä viisi </s> '
                         '<s> kuusi seitsemän kahdeksan </s> '
                         '<s> yhdeksän </s> '
                         '<s> kymmenen </s> '
                         '<s> kymmenen yhdeksän </s> '
                         '<s> kahdeksan seitsemän kuusi </s> '
                         '<s> viisi </s> '
                         '<s> neljä </s> '
                         '<s> kolme kaksi yksi </s>')
        assert_equal(all_file_ids, [0, 0, 0, 0,
                                    0, 0, 0, 0, 0,
                                    0, 0, 0, 0, 0,
                                    0, 0, 0,
                                    0, 0, 0,
                                    1, 1, 1, 1,
                                    1, 1, 1, 1, 1,
                                    1, 1, 1,
                                    1, 1, 1,
                                    1, 1, 1, 1, 1])

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import os
import logging
import subprocess
import numpy
import h5py
import theano
from theanolm import Vocabulary, Architecture, Network
from theanolm.scoring import LatticeDecoder, SLFLattice
from theanolm.filetypes import TextFileType

def add_arguments(parser):
    argument_group = parser.add_argument_group("files")
    argument_group.add_argument(
        'model_path', metavar='MODEL-FILE', type=str,
        help='the model file that will be used to decode the lattice')
    argument_group.add_argument(
        '--lattices', metavar='FILE', type=str, nargs='*', default=[],
        help='word lattices to be decoded (SLF, assumed to be compressed if '
             'the name ends in ".gz")')
    argument_group.add_argument(
        '--lattice-list', metavar='FILE', type=TextFileType('r'),
        help='text file containing a list of word lattices to be decoded (one '
             'path to an SLF file per line, the list and the SLF files are '
             'assumed to be compressed if the name ends in ".gz")')
    argument_group.add_argument(
        '--output-file', metavar='FILE', type=TextFileType('w'), default='-',
        help='where to write the best paths through the lattices (default '
             'stdout, will be compressed if the name ends in ".gz")')
    argument_group.add_argument(
        '--num-jobs', metavar='J', type=int, default=1,
        help='divide the set of lattice files into J distinct batches, and '
             'process only batch I')
    argument_group.add_argument(
        '--job', metavar='I', type=int, default=0,
        help='the index of the batch that this job should process, between 0 '
             'and J-1')

    argument_group = parser.add_argument_group("decoding")
    argument_group.add_argument(
        '--output', metavar='FORMAT', type=str, default='ref',
        help='format of the output, one of "ref" (default, utterance ID '
             'followed by words), "trn" (words followed by utterance ID in '
             'parentheses), "full" (utterance ID, acoustic score, language '
             'score, and number of words, followed by words)')
    argument_group.add_argument(
        '--n-best', metavar='N', type=int, default=1,
        help='print N best paths of each lattice (default 1)')
    argument_group.add_argument(
        '--nnlm-weight', metavar='LAMBDA', type=float, default=1.0,
        help="language model probabilities given by the model read from "
             "MODEL-FILE will be weighted by LAMBDA, when interpolating with "
             "the language model probabilities in the lattice (default is 1.0, "
             "meaning that the LM probabilities in the lattice will be "
             "ignored)")
    argument_group.add_argument(
        '--lm-scale', metavar='LMSCALE', type=float, default=None,
        help="scale language model log probabilities by LMSCALE when computing "
             "the total probability of a path (default is to use the LM scale "
             "specified in the lattice file, or 1.0 if not specified)")
    argument_group.add_argument(
        '--wi-penalty', metavar='WIP', type=float, default=None,
        help="penalize word insertion by adding WIP to the total log "
             "probability as many times as there are words in the path "
             "(without scaling WIP by LMSCALE)")
    argument_group.add_argument(
        '--log-base', metavar='B', type=int, default=None,
        help="convert output log probabilities to base B and WIP from base B "
             "(default is natural logarithm; this does not affect reading "
             "lattices, since they specify their internal log base)")
    argument_group.add_argument(
        '--unk-penalty', metavar='LOGPROB', type=float, default=None,
        help="if LOGPROB is zero, do not include <unk> tokens in perplexity "
             "computation; otherwise use constant LOGPROB as <unk> token score "
             "(default is to use the network to predict <unk> probability)")
    argument_group.add_argument(
        '--linear-interpolation', action="store_true",
        help="use linear interpolation of language model probabilities, "
             "instead of (pseudo) log-linear")

    argument_group = parser.add_argument_group("pruning")
    argument_group.add_argument(
        '--max-tokens-per-node', metavar='T', type=int, default=None,
        help="keep only at most T tokens at each node when decoding a lattice "
             "(default is no limit)")
    argument_group.add_argument(
        '--beam', metavar='B', type=float, default=None,
        help="prune tokens whose log probability is at least B smaller than "
             "the log probability of the best token at any given time (default "
             "is no beam pruning)")
    argument_group.add_argument(
        '--recombination-order', metavar='O', type=int, default=None,
        help="keep only the best token, when at least O previous words are "
             "identical (default is to recombine tokens only if the entire "
             "word history matches)")

    argument_group = parser.add_argument_group("logging and debugging")
    argument_group.add_argument(
        '--log-file', metavar='FILE', type=str, default='-',
        help='path where to write log file (default is standard output)')
    argument_group.add_argument(
        '--log-level', metavar='LEVEL', type=str, default='info',
        help='minimum level of events to log, one of "debug", "info", "warn" '
             '(default "info")')
    argument_group.add_argument(
        '--debug', action="store_true",
        help='enables debugging Theano errors')
    argument_group.add_argument(
        '--profile', action="store_true",
        help='enables profiling Theano functions')

def decode(args):
    log_file = args.log_file
    log_level = getattr(logging, args.log_level.upper(), None)
    if not isinstance(log_level, int):
        print("Invalid logging level requested:", args.log_level)
        sys.exit(1)
    log_format = '%(asctime)s %(funcName)s: %(message)s'
    if args.log_file == '-':
        logging.basicConfig(stream=sys.stdout, format=log_format, level=log_level)
    else:
        logging.basicConfig(filename=log_file, format=log_format, level=log_level)

    if args.debug:
        theano.config.compute_test_value = 'warn'
    else:
        theano.config.compute_test_value = 'off'
    theano.config.profile = args.profile
    theano.config.profile_memory = args.profile

    with h5py.File(args.model_path, 'r') as state:
        print("Reading vocabulary from network state.")
        sys.stdout.flush()
        vocabulary = Vocabulary.from_state(state)
        print("Number of words in vocabulary:", vocabulary.num_words())
        print("Number of word classes:", vocabulary.num_classes())
        print("Building neural network.")
        sys.stdout.flush()
        architecture = Architecture.from_state(state)
        network = Network(architecture, vocabulary,
                          mode=Network.Mode(minibatch=False))
        print("Restoring neural network state.")
        sys.stdout.flush()
        network.set_state(state)

    log_scale = 1.0 if args.log_base is None else numpy.log(args.log_base)

    if args.wi_penalty is None:
        wi_penalty = None
    else:
        wi_penalty = args.wi_penalty * log_scale
    if args.unk_penalty is None:
        ignore_unk = False
        unk_penalty = None
    elif args.unk_penalty == 0:
        ignore_unk = True
        unk_penalty = None
    else:
        ignore_unk = False
        unk_penalty = args.unk_penalty
    decoding_options = {
        'nnlm_weight': args.nnlm_weight,
        'lm_scale': args.lm_scale,
        'wi_penalty': wi_penalty,
        'ignore_unk': ignore_unk,
        'unk_penalty': unk_penalty,
        'linear_interpolation': args.linear_interpolation,
        'max_tokens_per_node': args.max_tokens_per_node,
        'beam': args.beam,
        'recombination_order': args.recombination_order
    }
    logging.debug("DECODING OPTIONS")
    for option_name, option_value in decoding_options.items():
        logging.debug("%s: %s", option_name, str(option_value))

    print("Building word lattice decoder.")
    sys.stdout.flush()
    decoder = LatticeDecoder(network, decoding_options)

    # Combine paths from command line and lattice list.
    lattices = args.lattices
    lattices.extend(args.lattice_list.readlines())
    lattices = [path.strip() for path in lattices]
    # Ignore empty lines in the lattice list.
    lattices = list(filter(None, lattices))
    # Pick every Ith lattice, if --num-jobs is specified and > 1.
    if args.num_jobs < 1:
        print("Invalid number of jobs specified:", args.num_jobs)
        sys.exit(1)
    if (args.job < 0) or (args.job > args.num_jobs - 1):
        print("Invalid job specified:", args.job)
        sys.exit(1)
    lattices = lattices[args.job::args.num_jobs]

    file_type = TextFileType('r')
    for index, path in enumerate(lattices):
        logging.info("Reading word lattice: %s", path)
        lattice_file = file_type(path)
        lattice = SLFLattice(lattice_file)

        if not lattice.utterance_id is None:
            utterance_id = lattice.utterance_id
        else:
            utterance_id = os.path.basename(lattice_file.name)
        logging.info("Utterance `%s' -- %d/%d of job %d",
                     utterance_id,
                     index + 1,
                     len(lattices),
                     args.job)
        tokens = decoder.decode(lattice)

        for index in range(min(args.n_best, len(tokens))):
            line = format_token(tokens[index],
                                utterance_id,
                                vocabulary,
                                log_scale,
                                args.output)
            args.output_file.write(line + "\n")

def format_token(token, utterance_id, vocabulary, log_scale, format):
    """Formats an output line from a token and an utterance ID.
    """

    words = vocabulary.id_to_word[token.history]
    if format == 'ref':
        return "{} {}".format(utterance_id, ' '.join(words))
    elif format == 'trn':
        return "{} ({})".format(' '.join(words), utterance_id)
    elif format == 'full':
        return "{} {} {} {} {}".format(
            utterance_id,
            token.ac_logprob / log_scale,
            token.lm_logprob / log_scale,
            len(words),
            ' '.join(words))
    else:
        print("Invalid output format requested:", args.output)
        sys.exit(1)



#!/usr/bin/env python3

import sys
from os import path
import subprocess
from glob import glob
import re
from setuptools import setup, find_packages

script_dir = path.dirname(path.realpath(__file__))
version_path = path.join(script_dir, 'theanolm', 'version.py')

# Don't import theanolm, as the user may not have the dependencies installed
# yet. This will import __version__.
with open(version_path, 'r') as version_file:
    exec(version_file.read())

long_description = 'TheanoLM is a recurrent neural network language modeling ' \
                   'toolkit implemented using Theano. Theano allows the user ' \
                   'to customize and extend the neural network very ' \
                   'conveniently, still generating highly efficient code ' \
                   'that can utilize multiple GPUs or CPUs for parallel ' \
                   'computation. TheanoLM allows the user to specify ' \
                   'arbitrary network architecture. New layer types and ' \
                   'optimization methods can be easily implemented.'
keywords = 'theano neural network language modeling machine learning research'
classifiers = ['Development Status :: 4 - Beta',
               'Environment :: Console',
               'Programming Language :: Python :: 3',
               'Intended Audience :: Science/Research',
               'License :: OSI Approved :: Apache Software License',
               'Operating System :: OS Independent',
               'Topic :: Scientific/Engineering']

setup(name='TheanoLM',
      version=__version__,
      author='Seppo Enarvi',
      author_email='seppo2016@marjaniemi.com',
      url='https://github.com/senarvi/theanolm',
      download_url='https://github.com/senarvi/theanolm/tarball/v' + __version__,
      description='Toolkit for neural network language modeling using Theano',
      long_description=long_description,
      license='Apache License, Version 2.0',
      keywords=keywords,
      classifiers=classifiers,
      packages=find_packages(exclude=['tests']),
      package_data={'theanolm': ['architectures/*.arch']},
      scripts=['bin/theanolm', 'bin/wctool'],
      install_requires=['numpy', 'scipy', 'Theano', 'h5py'],
      test_suite='tests')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest
from theanolm.training import Trainer

class DummyTrainer(object):
    pass

class TestTrainers(unittest.TestCase):
    def setUp(self):
        self.dummy_trainer = DummyTrainer()

    def tearDown(self):
        pass

    def test_is_scheduled(self):
        self.dummy_trainer._updates_per_epoch = 9
        self.dummy_trainer.update_number = 1
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 2))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 3))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 10))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 2, 4))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 2, 3))
        self.dummy_trainer.update_number = 2
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 2))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 3))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 10))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 2, 3))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 2, 2))
        self.dummy_trainer.update_number = 3
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 2))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 3))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 10))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 2, 2))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 2, 1))
        self.dummy_trainer.update_number = 4
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 2))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 3))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 10))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 2, 1))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 2, 0))
        self.dummy_trainer.update_number = 5
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 2))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 3))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 10))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 2, 1))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 2, 0))
        self.dummy_trainer.update_number = 6
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 2))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 3))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 10))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 2, 3))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 2, 2))
        self.dummy_trainer.update_number = 7
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 2))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 3))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 10))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 2, 2))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 2, 1))
        self.dummy_trainer.update_number = 8
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 2))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 3))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 10))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 2, 1))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 2, 0))
        self.dummy_trainer.update_number = 9
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 2))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 3))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 10))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 2, 1))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 2, 0))
        self.dummy_trainer.update_number = 10
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 2))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 3))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 10))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 2, 4))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 2, 3))
        self.dummy_trainer.update_number = 11
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 2))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 3))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 10))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 2, 3))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 2, 2))
        self.dummy_trainer.update_number = 12
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 2))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 3))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 10))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 2, 2))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 2, 1))

        self.dummy_trainer._updates_per_epoch = 8
        self.dummy_trainer.update_number = 1
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 3, 2))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 3, 1))
        self.dummy_trainer.update_number = 2
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 3, 1))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 3, 0))
        self.dummy_trainer.update_number = 3
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 3, 1))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 3, 0))
        self.dummy_trainer.update_number = 4
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 3, 2))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 3, 1))
        self.dummy_trainer.update_number = 5
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 3, 1))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 3, 0))
        self.dummy_trainer.update_number = 6
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 3, 1))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 3, 0))
        self.dummy_trainer.update_number = 7
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 3, 1))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 3, 0))
        self.dummy_trainer.update_number = 8
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 3, 1))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 3, 0))
        self.dummy_trainer.update_number = 9
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 3, 2))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 3, 1))
        self.dummy_trainer.update_number = 10
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 3, 1))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 3, 0))
        self.dummy_trainer.update_number = 11
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 3, 1))
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 3, 0))
        self.dummy_trainer.update_number = 12
        self.assertTrue(Trainer._is_scheduled(self.dummy_trainer, 3, 2))
        self.assertFalse(Trainer._is_scheduled(self.dummy_trainer, 3, 1))

if __name__ == '__main__':
    unittest.main()
